"""
Fit Parameters
--------------

This package uses functions from ``input_output``, ``equations_of_state``, and
``thermodynamics`` to fit parameters to experimental data.

input.json files have a different dictionary structure that is processed by
:func:`~despasito.input_output.read_input.process_param_fit_inputs`.
"""

import os
import numpy as np
from importlib import import_module
import logging

from . import fit_functions as ff
from . import data_classes

logger = logging.getLogger(__name__)


def fit(
    optimization_parameters=None,
    exp_data=None,
    global_opts={},
    minimizer_opts=None,
    MultiprocessingObject=None,
    **kwargs,
):
    r"""
    Fit parameters for an equation of state object with given experimental data.

    Each set of experimental data is converted to an object with the built in ability
    to evaluate its part of objective function.
    To add another type of supported experimental data, add a class to the
    fit_classes.py file.

    Parameters
    ----------
    optimization_parameters : dict, Optional, default=None
        Parameters used in global fitting algorithm.

        - fit_bead (str) - Name of bead whose parameters are being fit. Should be
        within bead_configuration.
        - fit_parameter_names (list[str]) - This list contains the name of the
        parameter being fit (e.g. epsilon). See EOS documentation for supported
        parameter names. Cross interaction parameter names should be composed of a
        parameter name followed by the interacting bead's name, separated by an
        underscore (e.g. epsilon_CO2).
        - parameters_guess (list[float]), Optional - Initial guess for all parameters
        being fit. If this is not provided, the Eos object provides a guesses based on
        the parameter types.
        - \*_bounds (list[float]), Optional - This list contains the minimum and
        maximum of the parameter from a parameter listed in fit_parameter_names,
        represented in place of the asterisk. See , :ref:`startfitting-label`, for
        more information.

    exp_data : dict, Optional, default=None
        This dictionary is made up of a dictionary for each data set that the
        parameters are fit to. Each dictionary is converted into an object and saved
        back to this structure before parameter fitting begins. Each key is an
        arbitrary string used to identify the data set and used later in reporting
        objective function values during the fitting process. See data type objects
        for more details.

        - data_class_type (str) - One of the supported data type objects to fit
        parameters, :ref:`data-types`.
        - eos_obj (obj) - Equation of state output that writes pressure, max density,
        fugacity coefficient, updates parameters, and evaluates parameter fitting
        objective function. See equation of state documentation for more details.

    global_opts : dict, Optional, default={}
        Method and keyword arguments used in global optimization method. See
        :func:`~despasito.parameter_fitting.fit_functions.global_minimization`.

        - method (str), Optional - default='differential_evolution', Global
        optimization method used to fit parameters. See
        :func:`~despasito.parameter_fitting.fit_functions.global_minimization`.
        - Additional options, specific to the global optimization method

    minimizer_opts : dict, Optional, default=None
        Dictionary used to define minimization type and the associated options.

        - method (str) - Method available to ``scipy.optimize.minimize``
        - options (dict) - This dictionary contains the keyword arguments available
        to the chosen method

    MultiprocessingObject : obj, Optional
        Multiprocessing object,
        :class:`~despasito.utils.parallelization.MultiprocessingJob`
    kwargs :
        Other keywords of instructions for thermodynamic calculations and parameter
        fitting.

    Returns
    -------
    output : dict
        Results from parameter optimization

        - parameters_final (numpy.ndarray) -  Array of the same length as
        `fit_parameter_names` containing the parameters resulting from the chosen
        global optimization method.
        - objective_value (float) - Objective value resulting from `parameters_final`

    """

    # Extract relevant quantities from kwargs
    dicts = {}

    # Extract inputs
    if optimization_parameters is None:
        raise ValueError("Required input, optimization_parameters, is missing.")

    dicts["global_opts"] = global_opts
    if minimizer_opts is not None:
        dicts["minimizer_opts"] = minimizer_opts

    if exp_data is None:
        raise ValueError("Required input, exp_data, is missing.")

    # Add multiprocessing object to exp_data objects and global_optss
    if MultiprocessingObject is not None:
        for k2 in list(exp_data.keys()):
            exp_data[k2]["MultiprocessingObject"] = MultiprocessingObject
        dicts["global_opts"]["MultiprocessingObject"] = MultiprocessingObject

    # Thermodynamic options and optimization options are added to data object
    for k2 in list(exp_data.keys()):
        for key, value in kwargs.items():
            if key not in exp_data[k2]:
                exp_data[k2][key] = value

    if not isinstance(optimization_parameters["fit_parameter_names"], list):
        if isinstance(optimization_parameters["fit_parameter_names"], str):
            optimization_parameters["fit_parameter_names"] = [optimization_parameters["fit_parameter_names"]]
        else:
            raise ValueError(
                f"'fit_parameter_names' must be a list not: {optimization_parameters['fit_parameter_names']}"
            )

    # Generate initial guess and bounds for parameters if none was given
    optimization_parameters = ff.consolidate_bounds(optimization_parameters).copy()
    if "bounds" in optimization_parameters:
        bounds = optimization_parameters["bounds"]
        del optimization_parameters["bounds"]
    else:
        bounds = np.zeros((len(optimization_parameters["fit_parameter_names"]), 2))
    Eos = exp_data[list(exp_data.keys())[0]][
        "eos_obj"
    ]  # since all exp data sets use the same Eos, it doesn't really matter
    bounds = ff.check_parameter_bounds(optimization_parameters, Eos, bounds)

    if "parameters_guess" in optimization_parameters:
        parameters_guess = optimization_parameters["parameters_guess"]
        if len(parameters_guess) != len(optimization_parameters["fit_parameter_names"]):
            raise ValueError("The number of initial parameters given isn't the same number of " "parameters to be fit.")
    else:
        parameters_guess = ff.initial_guess(optimization_parameters, Eos)
    logger.info("Initial guess in parameters: {}".format(parameters_guess))

    # _________________________________________________________

    # Reformat exp. data into formatted dictionary
    exp_dict = {}
    pkgpath = os.path.dirname(data_classes.__file__)
    type_list = [f for f in os.listdir(pkgpath) if f.endswith(".py")]
    type_list.remove("__init__.py")

    for key, data_dict in exp_data.items():
        fittype = data_dict["data_class_type"]
        try:
            exp_module = import_module("." + fittype, package="despasito.parameter_fitting.data_classes")
            data_class = getattr(exp_module, "Data")
        except Exception:
            if not type_list:
                raise ImportError("No fit types")
            elif len(type_list) == 1:
                tmp = type_list[0]
            else:
                tmp = ", ".join(type_list)
            raise ImportError(
                "The experimental data type, '{}', was not found\nThe following "
                "calculation types are supported: {}".format(fittype, tmp)
            )

        try:
            instance = data_class(data_dict)
            exp_dict[key] = instance
            logger.info("Initiated exp. data object: {}".format(instance.name))
        except Exception:
            raise AttributeError("Data set, {}, did not properly initiate object".format(key))

    # Check global optimization method
    if "method" in dicts["global_opts"]:
        global_method = dicts["global_opts"]["method"]
        del dicts["global_opts"]["method"]
    else:
        global_method = "differential_evolution"

    # Run Parameter Fitting
    try:
        result = ff.global_minimization(
            global_method,
            parameters_guess,
            bounds,
            optimization_parameters["fit_bead"],
            optimization_parameters["fit_parameter_names"],
            exp_dict,
            **dicts,
        )

        logger.info("Fitting terminated:\n{}".format(result.message))
        logger.info("Best Fit Parameters")
        logger.info("    Obj. Value: {}".format(result.fun))
        for i in range(len(optimization_parameters["fit_parameter_names"])):
            logger.info(
                "    {} {}: {}".format(
                    optimization_parameters["fit_bead"],
                    optimization_parameters["fit_parameter_names"][i],
                    result.x[i],
                )
            )

    except Exception:
        raise TypeError("The parameter fitting failed")

    return {
        "fit_bead": optimization_parameters["fit_bead"],
        "fit_parameter_names": optimization_parameters["fit_parameter_names"],
        "parameters_final": result.x,
        "objective_value": result.fun,
    }



""" Combining rules options called from
:func:`~despasito.equations_of_state.eos_toolbox.combining_rules` in the EOS class.
"""

import numpy as np
import logging

# Special functions are imported as needed

logger = logging.getLogger(__name__)


def mean(beadA, beadB, parameter):
    r"""
    Calculates cross interaction parameter with a mean: c = (a+b)/2

    Parameters
    ----------
    beadA : dict
        Dictionary of parameters used to describe a bead
    beadB : dict
        Dictionary of parameters used to describe a bead
    parameter : str
        Name of parameter for which a mixed value is needed

    Returns
    -------
    parameter12 : float
        Mixed interaction parameter

    """

    return (beadA[parameter] + beadB[parameter]) / 2


def geometric_mean(beadA, beadB, parameter):
    r"""
    Calculates cross interaction parameter with a geometric mean: c = np.sqrt(a*b)

    Parameters
    ----------
    beadA : dict
        Dictionary of parameters used to describe a bead
    beadB : dict
        Dictionary of parameters used to describe a bead
    parameter : str
        Name of parameter for which a mixed value is needed

    Returns
    -------
    parameter12 : float
        Mixed interaction parameter

    """

    return np.sqrt(beadA[parameter] * beadB[parameter])


def volumetric_geometric_mean(beadA, beadB, parameter, weighting_parameters=[]):
    r"""
    Calculates cross interaction parameter with a volumetric geometric mean:
        c = np.sqrt(a[0]*b[0]) * np.sqrt(a[1]**3 * b[1]**3) / ((a[1] + b[1])/2)**3

    Parameters
    ----------
    beadA : dict
        Dictionary of parameters used to describe a bead
    beadB : dict
        Dictionary of parameters used to describe a bead
    parameter : str
        Name of parameter for which a mixed value is needed
    weighting_parameters : list[str], Optional, default=[]
        Given parameter name is a0 and b0, while weighting_parameters should be of
        length 1 to represent the name for a1 and b1.

    Returns
    -------
    parameter12 : float
        Mixed interaction parameter

    """

    tmp1 = np.sqrt(beadA[parameter] * beadB[parameter])
    param2 = weighting_parameters[0]
    tmp2 = np.sqrt((beadA[param2] ** 3) * (beadB[param2] ** 3)) * 8 / ((beadA[param2] + beadB[param2]) ** 3)
    return tmp1 * tmp2


def weighted_mean(beadA, beadB, parameter, weighting_parameters=[]):
    r"""
    Calculates cross interaction parameter with a weighted mean:
        (a[0]*a[1] + b[0]*b[1]) / (a[1] + b[1])

    Parameters
    ----------
    beadA : dict
        Dictionary of parameters used to describe a bead
    beadB : dict
        Dictionary of parameters used to describe a bead
    parameter : str
        Name of parameter for which a mixed value is needed
    weighting_parameters : list[str], Optional, default=[]
        Given parameter name is a0 and b0, while weighting_parameters should be of
        length 1 to represent the name for a1 and b1.

    Returns
    -------
    parameter12 : float
        Mixed interaction parameter

    """

    param2 = weighting_parameters[0]
    parameter12 = (beadA[parameter] * beadA[param2] + beadB[parameter] * beadB[param2]) / (
        beadA[param2] + beadB[param2]
    )

    return parameter12


def mie_exponent(beadA, beadB, parameter):
    r"""
    Calculates cross interaction parameter like a mie_exponent:
        3 + np.sqrt((a-3)*(b-3))

    Parameters
    ----------
    beadA : dict
        Dictionary of parameters used to describe a bead
    beadB : dict
        Dictionary of parameters used to describe a bead
    parameter : str
        Name of parameter for which a mixed value is needed

    Returns
    -------
    parameter12 : float
        Mixed interaction parameter

    """

    return 3 + np.sqrt((beadA[parameter] - 3.0) * (beadB[parameter] - 3.0))


def square_well_berthelot(beadA, beadB, parameter, weighting_parameters=[]):
    r"""
    Calculates cross interaction parameter with a Square_well Berthelot geometric mean:
        c = np.sqrt(a[0]*b[0]) * np.sqrt(a[1]**3 * b[1]**3) / ((a[1] + b[1])/2)**3

    Parameters
    ----------
    beadA : dict
        Dictionary of parameters used to describe a bead
    beadB : dict
        Dictionary of parameters used to describe a bead
    parameter : str
        Name of parameter for which a mixed value is needed
    weighting_parameters : list[str], Optional, default=[]
        Given parameter name is a0 and b0, while weighting_parameters should be of
        length 1 to represent the name for a1 and b1.

    Returns
    -------
    parameter12 : float
        Mixed interaction parameter

    """

    param2, param3 = weighting_parameters[0], weighting_parameters[1]

    tmp1 = np.sqrt(beadA[parameter] * beadB[parameter])
    tmp2 = np.sqrt((beadA[param2] ** 3) * (beadB[param2] ** 3)) * 8 / ((beadA[param2] + beadB[param2]) ** 3)

    param3_12 = weighted_mean(beadA, beadB, param3, weighting_parameters=[param2])
    tmp3 = np.sqrt((beadA[param3] ** 3 - 1) * (beadB[param3] ** 3 - 1)) / (param3_12**3 - 1)

    return tmp1 * tmp2 * tmp3


def multipole(beadA, beadB, parameter, temperature=None, mode="curve fit", scaled=False):
    r"""
    Calculates cross interaction parameter with the multipole combining rules from
    the plug-in `MAPSCI <https://github.com/jaclark5/mapsci>`_.

    Parameters
    ----------
    beadA : dict
        Dictionary of parameters used to describe a bead
    beadB : dict
        Dictionary of parameters used to describe a bead
    parameter : str
        Name of parameter for which a mixed value is needed
    weighting_parameters : list[str], Optional, default=[]
        Given parameter name is a0 and b0, while weighting_parameters should be of
        length 1 to represent the name for a1 and b1.
    mode : str, Optional, default='curve fit'
        Dictates the mode by which the parameters are fit. By default the Mie
        parameters are fit to the multipole with the keyword "curve fit".
        Alternatively, the keyword "analytical" indicates that the energy parameter is
        explicitly calculated from the definite integral.
    scaled : bool, Optional, default=False
        Dictates whether the shape factor is used to scale the Mie potential

    Returns
    -------
    output : dict
        Mixed interaction parameter

    """

    try:
        import mapsci as mr
    except Exception:
        raise ImportError(
            "Multipole combining rules require 'mapsci' package, which is currently"
            " unavailable. Install it from: https://github.com/jaclark5/mapsci"
        )

    if scaled in [True, "True", "true", "yes", "Yes"]:
        shape_factor_scale = True
    else:
        shape_factor_scale = False

    if not isinstance(temperature, str) and temperature is None:
        tmp = {"beadA": beadA.copy(), "beadB": beadB.copy()}
        for key, value in tmp.items():
            tmp[key]["sigma"] = value["sigma"] * 10  # convert from nm to angstroms

        if mode == "curve fit":
            dict_cross, _ = mr.extended_combining_rules_fitting(tmp, temperature, shape_factor_scale=shape_factor_scale)
        elif mode == "analytical":
            dict_cross, _ = mr.extended_combining_rules_analytical(
                tmp, temperature, shape_factor_scale=shape_factor_scale
            )
        else:
            raise ValueError("Multipole mixing rule must be either 'curve fit' or 'analytical'.")
        output = dict_cross["beadA"]["beadB"]
    else:
        logger.warning("Temperature is None, using geometric mean.")
        output = {parameter: geometric_mean(beadA, beadB, parameter)}

    return output



"""
Unit and regression test for the despasito package.
"""

# Import package, test suite, and other packages as needed
import despasito.thermodynamics as thermo
import despasito.equations_of_state
import pytest
import sys
import numpy as np

Tlist = [353.0]
xilist = [[0.002065, 0.997935]]
yilist = [[0.98779049, 0.01220951]]
Plist = [7000000.0]
rho_co2_h2o = np.array([19986.78358869])

bead_library = {
    "H2O353": {
        "epsilon": 479.56,
        "lambdaa": 6.0,
        "lambdar": 8.0,
        "sigma": 3.0029e-1,
        "Sk": 1.0,
        "Vks": 1,
        "mass": 0.018015,
    },
    "CO2": {
        "epsilon": 353.55,
        "lambdaa": 6.66,
        "lambdar": 23.0,
        "sigma": 3.741e-1,
        "Sk": 1.0,
        "Vks": 1,
        "mass": 0.04401,
    },
    "CH3CH2CH2-": {
        "Sk": 1.0,
        "Vks": 1.0,
        "epsilon": 342.0,
        "lambdaa": 6.0,
        "lambdar": 15.0,
        "mass": 0.04309,
        "sigma": 0.45089,
    },
}
cross_library = {
    "CO2": {"H2O353": {"epsilon": 432.69}},
    "H2O353": {"CH3CH2CH2-": {"epsilon": 315.7604284}},
}

beads_co2_h2o = ["CO2", "H2O353"]
molecular_composition_co2_h2o = np.array([[1.0, 0.0], [0.0, 1.0]])
Eos_co2_h2o = despasito.equations_of_state.initiate_eos(
    eos="saft.gamma_mie",
    beads=beads_co2_h2o,
    molecular_composition=molecular_composition_co2_h2o,
    bead_library=bead_library,
    cross_library=cross_library,
    numba=True,
)

beads_co2_h2o = ["H2O353", "CH3CH2CH2-"]
molecular_composition_co2_h2o = np.array([[1.0, 0.0], [0.0, 2.0]])
Eos_h2o_hexane = despasito.equations_of_state.initiate_eos(
    eos="saft.gamma_mie",
    beads=beads_co2_h2o,
    molecular_composition=molecular_composition_co2_h2o,
    bead_library=bead_library,
    cross_library=cross_library,
)


def test_thermo_import():
    #    """Sample test, will always pass so long as import statement worked"""
    assert "despasito.thermodynamics" in sys.modules


def test_saturation_properties(Eos=Eos_co2_h2o, Tlist=Tlist):

    output = thermo.thermo(
        Eos, calculation_type="saturation_properties", **{"Tlist": Tlist, "xilist": [np.array([0.0, 1.0])]}
    )

    assert output["Psat"][0] == pytest.approx(46266.2, abs=1e1) and output["rhol"][0] == pytest.approx(
        53883.63, abs=1e-1
    ), output["rhol"][0] == pytest.approx(2371.38970066, abs=1e-1)


def test_liquid_properties(Eos=Eos_co2_h2o, Tlist=Tlist, xilist=xilist, Plist=Plist):

    output = thermo.thermo(
        Eos, calculation_type="liquid_properties", **{"Tlist": Tlist, "Plist": Plist, "xilist": xilist}
    )

    assert output["rhol"][0] == pytest.approx(53831.6, abs=1e-1) and output["phil"][0] == pytest.approx(
        np.array([403.98, 6.8846e-03]), abs=1e-1
    )


def test_vapor_properties(Eos=Eos_co2_h2o, Tlist=Tlist, yilist=yilist, Plist=Plist):

    output = thermo.thermo(
        Eos, calculation_type="vapor_properties", **{"Tlist": Tlist, "Plist": Plist, "yilist": yilist}
    )

    assert output["rhov"][0] == pytest.approx(2938.3, abs=1e-1) and output["phiv"][0] == pytest.approx(
        np.array([0.865397, 0.63848]), abs=1e-1
    )


def test_activity_coefficient(Eos=Eos_h2o_hexane, Tlist=Tlist, xilist=xilist, yilist=yilist, Plist=Plist):

    output = thermo.thermo(
        Eos,
        calculation_type="activity_coefficient",
        **{"Tlist": Tlist, "Plist": Plist, "yilist": yilist, "xilist": xilist},
    )

    print(output["gamma"])
    assert output["gamma"][0] == pytest.approx(np.array([7.23733364e04, 6.30243983e-01]), abs=1e-2)


def test_bubble_pressure(Eos=Eos_co2_h2o, Tlist=Tlist, xilist=xilist):
    output = thermo.thermo(
        Eos,
        calculation_type="bubble_pressure",
        **{"Tlist": Tlist, "xilist": xilist, "Pmin": [6900000], "Pmax": [7100000]},
    )
    assert output["P"][0] == pytest.approx(7005198.6, abs=5e1) and output["yi"][0] == pytest.approx(
        [0.98779049, 0.01220951], abs=1e-4
    )



"""
Unit and regression test for the despasito package.
"""

# Import package, test suite, and other packages as needed
import copy
import pytest
import sys
import numpy as np

import despasito.equations_of_state

if "cython" not in sys.modules:
    print("Cython package is unavailable, using Numba")
    flag_cython = False
else:
    flag_cython = True

path = "despasito.equations_of_state.saft.compiled_modules"

xi_co2_ben = np.array([0.2, 0.2])
beads_co2_ben = ["CO2", "benzene"]
molecular_composition_co2_ben = np.array([[1.0, 0.0], [0.0, 1.0]])
bead_library_co2_ben = {
    "CO2": {
        "epsilon": 361.69,
        "lambdaa": 6.66,
        "lambdar": 23.0,
        "sigma": 3.741e-1,
        "Sk": 1.0,
        "Vks": 1.0,
        "mass": 0.04401,
    },
    "baCH": {
        "epsilon": 371.53,
        "lambdaa": 6.0,
        "lambdar": 14.756,
        "sigma": 4.0578e-1,
        "Sk": 0.32184,
        "Vks": 1.0,
        "mass": 0.01302,
    },
    "baCHCH": {
        "epsilon": 243.88,
        "lambdaa": 6.0,
        "lambdar": 11.58,
        "sigma": 3.482e-1,
        "Sk": 1.0,
        "Vks": 1.0,
        "mass": 0.02604,
    },
    "baCHCHCH": {
        "epsilon": 353.93,
        "lambdaa": 6.0,
        "lambdar": 14.23,
        "sigma": 3.978e-1,
        "Sk": 1.0,
        "Vks": 1.0,
        "mass": 0.03905,
    },
    "benzene": {
        "epsilon": 658.17,
        "lambdaa": 6.0,
        "lambdar": 32.0,
        "sigma": 3.842e-1,
        "Sk": 1.0,
        "Vks": 1.0,
        "mass": 0.07811,
    },
}

xi_co2_h2o = np.array([0.78988277, 0.21011723])
beads_co2_h2o = ["CO2", "H2O"]
molecular_composition_co2_h2o = np.array([[1.0, 0.0], [0.0, 1.0]])
bead_library_co2_h2o = {
    "H2O": {
        "epsilon": 266.68,
        "lambdaa": 6.0,
        "lambdar": 17.02,
        "sigma": 3.0063e-1,
        "Sk": 1.0,
        "Vks": 1,
        "mass": 0.018015,
        "Nk-H": 2,
        "Nk-e1": 2,
        "epsilonHB-H-e1": 1985.4,
        "K-H-e1": 1.0169e-1,
    },
    "CO2": {
        "epsilon": 207.89,
        "lambdaa": 5.055,
        "lambdar": 26.408,
        "sigma": 3.05e-1,
        "Sk": 0.8468,
        "Vks": 2,
        "mass": 0.04401,
        "Nk-H": 1,
        "Nk-a1": 1,
    },
}
cross_library_co2_h2o = {"CO2": {"H2O": {"epsilon": 226.38, "epsilonHB-H-e1": 2200.0, "K-H-e1": 9.1419e-2}}}
epsilonHB_co2_h2o = np.array(
    [
        [
            [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
            [[0.0, 0.0, 2200.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
        ],
        [
            [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [2200.0, 0.0, 0.0]],
            [[0.0, 0.0, 1985.4], [0.0, 0.0, 0.0], [1985.4, 0.0, 0.0]],
        ],
    ]
)

T = 323.2
rho_co2_h2o = np.array([21146.16997993])
P = np.array([15727315.77])


def test_saft_gamma_mie_imported():
    #    """Sample test, will always pass so long as import statement worked"""
    assert "despasito.equations_of_state" in sys.modules


def test_saft_gamma_mie_class_noassoc(
    beads=beads_co2_ben,
    molecular_composition=molecular_composition_co2_ben,
    bead_library=bead_library_co2_ben,
):
    #   """Test ability to create EOS object without association sites"""
    Eos_class = despasito.equations_of_state.initiate_eos(
        eos="saft.gamma_mie",
        beads=beads,
        molecular_composition=molecular_composition,
        bead_library=copy.deepcopy(bead_library),
    )
    assert (Eos_class.eos_dict["massi"] == np.array([0.04401, 0.07811])).all()


def test_saft_gamma_mie_class_assoc(
    beads=beads_co2_h2o,
    molecular_composition=molecular_composition_co2_h2o,
    bead_library=bead_library_co2_h2o,
    cross_library=cross_library_co2_h2o,
    epsilonHB=epsilonHB_co2_h2o,
):
    #   """Test ability to create EOS object with association sites"""
    Eos_class = despasito.equations_of_state.initiate_eos(
        eos="saft.gamma_mie",
        beads=beads,
        molecular_composition=molecular_composition,
        bead_library=copy.deepcopy(bead_library),
        cross_library=copy.deepcopy(cross_library),
    )
    assert (Eos_class.eos_dict["epsilonHB"] == epsilonHB).all()


def test_saft_gamma_mie_class_assoc_P(
    T=T,
    xi=xi_co2_h2o,
    rho=rho_co2_h2o,
    beads=beads_co2_h2o,
    molecular_composition=molecular_composition_co2_h2o,
    bead_library=bead_library_co2_h2o,
    cross_library=cross_library_co2_h2o,
):
    #   """Test ability to create EOS object with association sites"""
    Eos_class = despasito.equations_of_state.initiate_eos(
        eos="saft.gamma_mie",
        beads=beads,
        molecular_composition=molecular_composition,
        bead_library=copy.deepcopy(bead_library),
        cross_library=copy.deepcopy(cross_library),
    )
    #   """Test ability to predict P with association sites"""
    P = Eos_class.pressure(rho, T, xi)[0]

    assert P == pytest.approx(15727315.77, abs=1e3)


def test_saft_gamma_mie_class_assoc_fugacity_coeff(
    P=P,
    T=T,
    xi=xi_co2_h2o,
    rho=rho_co2_h2o,
    beads=beads_co2_h2o,
    molecular_composition=molecular_composition_co2_h2o,
    bead_library=bead_library_co2_h2o,
    cross_library=cross_library_co2_h2o,
):
    #   """Test ability to create EOS object with association sites"""
    Eos_class = despasito.equations_of_state.initiate_eos(
        eos="saft.gamma_mie",
        beads=beads,
        molecular_composition=molecular_composition,
        bead_library=copy.deepcopy(bead_library),
        cross_library=copy.deepcopy(cross_library),
    )
    #   """Test ability to predict P with association sites"""
    phi = Eos_class.fugacity_coefficient(P, rho, xi, T)
    assert phi == pytest.approx(np.array([0.48972481, 0.00281112]), abs=1e-4)


def test_numba_available():
    assert path + ".ext_Aassoc_numba" in sys.modules and path + ".ext_gamma_mie_numba" in sys.modules


def test_saft_gamma_mie_class_assoc_P_numba(
    T=T,
    xi=xi_co2_h2o,
    rho=rho_co2_h2o,
    beads=beads_co2_h2o,
    molecular_composition=molecular_composition_co2_h2o,
    bead_library=bead_library_co2_h2o,
    cross_library=cross_library_co2_h2o,
):
    #   """Test ability to predict P with association sites"""
    Eos = despasito.equations_of_state.initiate_eos(
        eos="saft.gamma_mie",
        beads=beads,
        molecular_composition=molecular_composition,
        bead_library=copy.deepcopy(bead_library),
        cross_library=copy.deepcopy(cross_library),
        numba=True,
    )

    #   """Test ability to predict P with association sites"""
    P = Eos.pressure(rho, T, xi)[0]

    assert P == pytest.approx(15727315.77, abs=1e3)


@pytest.mark.skipif(not flag_cython, reason="Cython is not installed with this version of python.")
def test_cython_available():
    assert path + ".ext_Aassoc_cython" in sys.modules and path + ".ext_gamma_mie_cython" in sys.modules


@pytest.mark.skipif(not flag_cython, reason="Cython is not installed with this version of python.")
def test_saft_gamma_mie_class_assoc_P_cython(
    T=T,
    xi=xi_co2_h2o,
    rho=rho_co2_h2o,
    beads=beads_co2_h2o,
    molecular_composition=molecular_composition_co2_h2o,
    bead_library=bead_library_co2_h2o,
    cross_library=cross_library_co2_h2o,
):
    #   """Test ability to predict P with association sites"""
    Eos = despasito.equations_of_state.initiate_eos(
        eos="saft.gamma_mie",
        beads=beads,
        molecular_composition=molecular_composition,
        bead_library=copy.deepcopy(bead_library),
        cross_library=copy.deepcopy(cross_library),
        cython=True,
    )
    P = Eos.pressure(rho, T, xi)[0]
    assert P == pytest.approx(15727315.77, abs=1e3)



"""
Unit and regression test for the despasito package.
"""

# Import package, test suite, and other packages as needed
import despasito.equations_of_state

# import despasito.equations_of_state.saft.solv_assoc as solv_assoc
import copy

import pytest
import numpy as np

bead = ["H2O"]
molecular_composition = np.array([[1.0]])
bead_library = {
    "H2O": {
        "epsilon": 250.0,
        "lambda": 1.7889,
        "sigma": 3.0342e-1,
        "Sk": 1.0,
        "Vks": 1.0,
        "Nk-e": 2,
        "Nk-H": 2,
        "epsilonHB-e-H": 1400.0,
        "rc-e-H": 0.210822,
        "mass": 0.018015,
    }
}

Kklab = np.array([[[[[[0.0, 0.00106673], [0.00106673, 0.0]]]]]])

Eos = despasito.equations_of_state.initiate_eos(
    eos="saft.gamma_sw",
    beads=bead,
    molecular_composition=molecular_composition,
    bead_library=copy.deepcopy(bead_library),
)
T = 580.0
density = np.array([40004.84463113798])
P = np.array([9447510.360679299])


def test_saft_gamma_sw_class_bonding_volume(
    T=T,
    beads=bead,
    molecular_composition=molecular_composition,
    bead_library=bead_library,
    Kklab=Kklab,
):
    #   """Test ability to create EOS object with association sites"""
    Eos_class = despasito.equations_of_state.initiate_eos(
        eos="saft.gamma_sw",
        beads=beads,
        molecular_composition=molecular_composition,
        bead_library=copy.deepcopy(bead_library),
    )

    rc_klab = Eos_class.eos_dict["rc_klab"]
    Kklab_new = Eos_class.saft_source.calc_Kijklab(T, rc_klab)

    assert Kklab_new == pytest.approx(Kklab, abs=1e-7)


def test_saft_gamma_sw_class_assoc_P(T=T, xi=[1.0], Eos=Eos, density=density):
    #   """Test ability to predict P with association sites"""
    P = Eos.pressure(density, T, xi)[0]
    assert P == pytest.approx(9447510.360679299, abs=1e3)


def test_saft_gamma_sw_class_assoc_fugacity_coefficient(P=P, xi=[1.0], T=T, Eos=Eos, density=density):
    #   """Test ability to predict P with association sites"""
    phi = Eos.fugacity_coefficient(P, density, xi, T)
    assert phi == pytest.approx(np.array([0.8293442]), abs=1e-4)


Eos = despasito.equations_of_state.initiate_eos(
    eos="saft.gamma_sw",
    beads=bead,
    molecular_composition=molecular_composition,
    bead_library=copy.deepcopy(bead_library),
    numba=True,
)


def test_saft_gamma_sw_class_assoc_P_numba(T=T, xi=np.array([1.0]), Eos=Eos, density=density):
    #   """Test ability to predict P with association sites"""
    P = Eos.pressure(density, T, xi)[0]
    assert P == pytest.approx(9447510.360679299, abs=1e3)



""" Functions needed for parameter fitting process.
"""

import numpy as np
import logging
from inspect import getmembers, isfunction
from scipy.optimize import NonlinearConstraint, LinearConstraint

from . import constraint_types as constraints_mod
from . import global_methods as global_methods_mod
import despasito.utils.general_toolbox as gtb

logger = logging.getLogger(__name__)


def initial_guess(optimization_parameters, Eos):
    r"""
    Extract initial guess in fit parameters from Eos object.

    Parameters
    ----------
    optimization_parameters : dict
        Parameters used in global optimization algorithm.

        - fit_bead (str) - Name of bead whose parameters are being fit.
        - fit_parameter_names (list[str]) - This list contains the name of the
        parameter being fit (e.g. epsilon). See EOS documentation for supported
        parameter names. Cross interaction parameter names should be composed of
        parameter name and the other bead type, separated by an underscore (e.g.
        epsilon_CO2).

    Eos : obj
        Equation of state object that writes pressure, max density, fugacity
        coefficient, updates parameters, and evaluates objective functions. For
        parameter fitting algorithm See equation of state documentation for more
        details.

    Returns
    -------
    parameters_guess : numpy.ndarray,
        An array of initial guesses for parameters to be optimized throughout the
        process.

    """

    # Update bead_library with test parameters

    parameters_guess = np.ones(len(optimization_parameters["fit_parameter_names"]))
    for i, param in enumerate(optimization_parameters["fit_parameter_names"]):
        fit_parameter_names_list = param.split("_")
        if len(fit_parameter_names_list) == 1:
            parameters_guess[i] = Eos.guess_parameters(
                fit_parameter_names_list[0], [optimization_parameters["fit_bead"]]
            )
        elif len(fit_parameter_names_list) == 2:
            parameters_guess[i] = Eos.guess_parameters(
                fit_parameter_names_list[0],
                [optimization_parameters["fit_bead"], fit_parameter_names_list[1]],
            )
        else:
            raise ValueError(
                "Parameters for only one bead are allowed to be fit. Multiple "
                + "underscores in a parameter name suggest more than one bead type "
                + "in your fit parameter name, {}".format(param)
            )

    return parameters_guess


def check_parameter_bounds(optimization_parameters, Eos, bounds):
    r"""
    Check that provided parameter bounds are within reasonable limits defined in Eos
    object.

    Parameters
    ----------
    optimization_parameters : dict
        Parameters used in global optimization algorithm

        - fit_parameter_names (list[str]) - This list contains the name of the
        parameter being fit (e.g. epsilon). See EOS documentation for supported
        parameter names. Cross interaction parameter names should be composed of
        parameter name and the other bead name separated by an underscore (e.g.
        epsilon_CO2).

    Eos : obj
        Equation of state object that writes pressure, max density, fugacity
        coefficient, updates parameters, and evaluates parameter fitting objective
        function. See equation of state documentation for more details.
    bounds : numpy.ndarray
        List of length equal to fit_parameter_names with lists of pairs for minimum
        and maximum bounds of parameter being fit. Defaults defined in Eos object are
        broad, so we recommend specification.

    Returns
    -------
    new_bounds : list[tuple]
        Checked with Eos object method, this list has a length equal to
        fit_parameter_names with lists of pairs for minimum and maximum bounds of
        parameter being fit. Bounds are adjusted to remain within limits set by Eos
        object.

    """

    new_bounds = [(0, 0) for x in range(len(optimization_parameters["fit_parameter_names"]))]
    # Check boundary parameters to be sure they're in a reasonable range
    for i, param in enumerate(optimization_parameters["fit_parameter_names"]):
        fit_parameter_names_list = param.split("_")
        new_bounds[i] = tuple(Eos.check_bounds(fit_parameter_names_list[0], param, bounds[i]))

    return new_bounds


def consolidate_bounds(optimization_parameters):
    r"""
    Parse parameter bounds in the ``optimization_parameters`` dictionary.

    The resulting bounds form a 2D numpy array with a length equal to the number of
    parameters being fit.

    Parameters
    ----------
    optimization_parameters : dict
        Parameters used in basin fitting algorithm

        - fit_bead (str) - Name of bead whose parameters are being fit, should be in
        bead list of bead_configuration
        - fit_parameter_names (list[str]) - This list contains the name of the
        parameter being fit (e.g. epsilon). See Eos object documentation for supported
        parameter names. Cross interaction parameter names should be composed of
        parameter name and the other bead type, separated by an underscore (e.g.
        epsilon_CO2).
        - \*_bounds (list[float]), Optional - This list contains the minimum and
        maximum of the parameter from a parameter listed in fit_parameter_names,
        represented in place of the asterisk. See , :ref:`startfitting-label`, for
        more information.

    Returns
    -------
    new_optimization_parameters : dict
        Parameters used in basin fitting algorithm

        - fit_bead (str) - Name of bead whose parameters are being fit.
        - fit_parameter_names (list[str]) - This list contains the name of the
        parameter being fit (e.g. epsilon). See EOS documentation for supported
        parameter names. Cross interaction parameter names should be composed of
        parameter name and the other bead type, separated by an underscore (e.g.
        epsilon_CO2).
        - bounds (numpy.ndarray) - List of lists of length two, of length equal to
        ``fit_parameter_names``. If no bounds were given then the default parameter
        boundaries are [0,1e+4], else bounds given as \*_bounds in input dictionary
        are used.

    """

    if "fit_bead" not in optimization_parameters:
        raise ValueError(
            "optimization_parameters dictionary should include keyword, fit_bead, "
            "defining the name of the bead whose parameters are to be fit."
        )
    if "fit_parameter_names" not in optimization_parameters:
        raise ValueError(
            "optimization_parameters dictionary should include keyword, "
            "fit_parameter_names, defining the parameters to be fit."
        )

    new_optimization_parameters = {}
    new_optimization_parameters["bounds"] = [
        [0, 1e4] for x in range(len(optimization_parameters["fit_parameter_names"]))
    ]
    for key2, value2 in optimization_parameters.items():
        if "bounds" in key2:
            tmp = key2.replace("_bounds", "")
            if tmp in optimization_parameters["fit_parameter_names"]:
                logger.info("Accepted bounds for parameter, '{}': {}".format(tmp, value2))
                ind = optimization_parameters["fit_parameter_names"].index(tmp)
                new_optimization_parameters["bounds"][ind] = value2
            else:
                logger.error(
                    "Bounds for parameter type '{}' were given, but this parameter is"
                    " not defined to be fit.".format(tmp)
                )
        else:
            new_optimization_parameters[key2] = value2
            continue

    return new_optimization_parameters


def reformat_output(cluster):
    r"""
    Takes a list of lists that contain thermo output of lists and floats and reformats"
    " it into a 2D numpy array.

    Parameters
    ----------
    cluster : list[list[list/floats]]
        A list of lists, where the inner list is made up of lists and floats

    Returns
    -------
    matrix : numpy.ndarray
        A 2D matrix
    len_cluster : list
        a list of lengths for each of the columns (whether 1 for float, or len(list))

    """

    # if input is a list or array
    if len(cluster) == 1:
        matrix = np.transpose(np.array(cluster[0]))
        if not gtb.isiterable(cluster[0]):
            len_cluster = [1]
        else:
            len_cluster = [len(cluster[0])]
    # If list of lists or arrays
    else:

        # Obtain dimensions of final matrix
        len_cluster = []
        for i, tmp_cluster in enumerate(cluster):
            if gtb.isiterable(tmp_cluster[0]):
                len_cluster.append(len(tmp_cluster[0]))
            else:
                len_cluster.append(1)
        matrix_tmp = np.zeros([len(cluster[0]), sum(len_cluster)])

        # Transfer information to final matrix
        ind = 0
        for i, val in enumerate(cluster):
            try:
                matrix = np.zeros([len(val[0]), len(val)])
            except Exception:
                matrix = np.zeros([1, len(val)])
            for j, tmp in enumerate(val):
                matrix[:, j] = tmp
            l = len_cluster[i]
            if l == 1:
                matrix_tmp[:, ind] = np.array(matrix)
                ind += 1
            else:
                if len(matrix) == 1:
                    matrix = matrix[0]

                for j in range(l):
                    matrix_tmp[:, ind] = matrix[j]
                    ind += 1

        matrix = np.array(matrix_tmp)

    return matrix, len_cluster


def global_minimization(global_method, *args, **kwargs):
    r"""
    Fit defined parameters for equation of state object with given experimental data.

    Each set of experimental data is converted to an object with the build in ability
    to evaluate its part of objective function.
    To add another type of supported experimental data, see :ref:`contribute-fitting`.

    Parameters
    ----------
    global_method : str
        Global optimization method used to fit parameters. See supported
        :mod:`~despasito.parameter_fitting.global_methods`.
    parameters_guess : numpy.ndarray,
        An array of initial guesses for parameters, these will be optimized throughout
        the process.
    bounds : list[tuple]
        List of length equal to fit_parameter_names with lists of pairs for minimum
        and maximum bounds of parameter being fit. Defaults are broad, recommend
        specification.
    fit_bead : str
        Name of bead whose parameters are being fit, should be in bead list of
        bead_configuration
    fit_parameter_names : list[str]
        This list contains the name of the parameter being fit (e.g. epsilon). See Eos
        object documentation for supported parameter names. Cross interaction
        parameter names should be composed of parameter name and the other bead type,
        separated by an underscore (e.g. epsilon_CO2).
    exp_dict : dict
        Dictionary of experimental data objects.
    global_opts : dict, Optional
        Keyword arguments of global optimization algorithm. See specific options in
        :mod:`~despasito.parameter_fitting.global_methods`. Note that unless in the
        keyword, ``workers`` is provided, the thermodynamic calculations will we split
        among the cores. Check the global optimization method to ensure it uses the
        ``workers`` keyword.
    minimizer_opts : dict, Optional
        Dictionary used to define minimization type and the associated options.

        - method (str) - Method available to scipy.optimize.minimize
        - options (dict) - This dictionary contains the kwargs available to the chosen
        method

    constraints : dict, Optional
        This dictionary of constraint types and their arguments will be converted into
        the appropriate form for the chosen optimization method.

    Returns
    -------
    Objective : float
        Sum of objective values according to appropriate weights. Output file saved in
        current working directory.

    """

    logger.info("Using global optimization method: {}".format(global_method))
    calc_list = [o[0] for o in getmembers(global_methods_mod) if (isfunction(o[1]) and o[0][0] != "_")]
    try:
        func = getattr(global_methods_mod, global_method)
    except Exception:
        raise ImportError(
            "The global minimization type, '{}',".format(global_method)
            + " was not found\nThe following "
            + "calculation types are supported: {}".format(", ".join(calc_list))
        )

    output = func(*args, **kwargs)

    return output


def initialize_constraints(constraints, constraint_type):
    r"""
    A tuple of either constraint classes or dictionaries as required by
    :mod:`~despasito.parameter_fitting.global_methods`.

    Parameters
    ----------
    constraints : dict
        This dictionary of constraint types and their arguments will be converted into
        the appropriate form for the chosen optimization method. Although the key can
        be anything, it must represent a dictionary containing:

        * function (str) - must be found in the dictionary and represent a valid
        function name from :mod:`~despasito.parameter_fitting.constraint_types`
        * type - entries depends on ``constraint_type``
        * args - Inputs into the functions (keys)

    constraint_type : str
        Either 'dict' or 'class'. Changes the constraint to the specified form.

        * 'dict': Allowed types, "eq" or "ineq", eq means must be zero, ineq means it
        must be non-negative
        * 'class': Allowed types, "nonlinear" or "linear", a keyword argument may also
        be added for the constraint class

    Returns
    -------
    new_constraints : tuple
        A tuple of either constraint classes or dictionaries as required by global
        optimization methods

    """

    calc_list = [o[0] for o in getmembers(constraints_mod) if (isfunction(o[1]) and o[0][0] != "_")]

    new_constraints = []
    for const_type, kwargs in constraints.items():

        if "function" not in kwargs:
            raise ValueError("Constraint function type is not included")

        try:
            func = getattr(constraints_mod, kwargs["function"])
        except Exception:
            raise ImportError(
                "The constraint type, '{}', was not".format(kwargs["function"])
                + " found\nThe following types are "
                + "supported: {}".format(", ".join(calc_list))
            )

        if "args" not in kwargs:
            raise ValueError("Constraint function, {}, is missing arguments".format(kwargs["function"]))

        if constraint_type == "class":
            if "type" not in kwargs or kwargs["type"] in ["linear", "nonlinear"]:
                raise ValueError(
                    "Constraint, {}, does not have type. ".format(kwargs["function"])
                    + "Type can be 'linear' or 'nonlinear'."
                )
            if kwargs["type"] == "linear":
                if "kwargs" not in kwargs:
                    output = LinearConstraint(func, kwargs["args"][0], kwargs["args"][1])
                else:
                    output = LinearConstraint(func, kwargs["args"][0], kwargs["args"][1], **kwargs)
            elif kwargs["type"] == "nonlinear":
                if "kwargs" not in kwargs:
                    output = NonlinearConstraint(func, kwargs["args"][0], kwargs["args"][1])
                else:
                    output = NonlinearConstraint(func, kwargs["args"][0], kwargs["args"][1], **kwargs)

        elif constraint_type == "dict":
            if "type" not in kwargs or kwargs["type"] in ["eq", "ineq"]:
                raise ValueError(
                    "Constraint, {}, does not".format(kwargs["function"]) + " have type. Type can be 'eq' or 'ineq'."
                )
            output = {"type": kwargs["type"], "function": func, "args": kwargs["args"]}
        else:
            raise ValueError("Constraint type {}, must be either 'class' or 'dict'.")

        new_constraints.append(output)

    return tuple(new_constraints)


def compute_obj(beadparams, fit_bead, fit_parameter_names, exp_dict, bounds, frozen_parameters=None):
    r"""
    Fit defined parameters for equation of state object with given experimental data.

    Each set of experimental data is converted to an object with the built-in ability
    to evaluate its part of objective function.
    To add another type of supported experimental data, see :ref:`contribute-fitting`.

    Parameters
    ----------
    parameters_guess : numpy.ndarray,
        An array of initial guesses for parameters, these will be optimized throughout
        the process.
    fit_bead : str
        Name of bead whose parameters are being fit.
    fit_parameter_names : list[str]
        This list contains the name of the parameter being fit (e.g. epsilon). See EOS
        documentation for supported parameter names. Cross interaction parameter names
        should be composed of parameter name and the other bead type, separated by an
        underscore (e.g. epsilon_CO2).
    exp_dict : dict
        Dictionary of experimental data objects.
    bounds : list[tuple]
        List of length equal to fit_parameter_names with lists of pairs containing
        minimum and maximum bounds of parameters being fit. Defaults from Eos object
        are broad, so we recommend specification.
    frozen_parameters : numpy.ndarray, Optional, default=None
        List of first parameters in the fit_parameter_names list that are frozen during
        minimization. This feature is currently used in the
        :func:`~despasito.parameter_fitting.global_methods.grid_minimization` method,
        enabled with the ``split_grid_minimization`` feature.

    Returns
    -------
    Objective : float
        Sum of objective values according to appropriate weights.

    """

    # Update bead_library with test parameters

    if len(beadparams) != len(fit_parameter_names):
        if frozen_parameters is not None:
            beadparams = np.array(list(frozen_parameters) + list(beadparams))
        else:
            raise ValueError(
                "The length of initial guess vector should be the same number of " + "parameters to be fit."
            )

    logger.info(
        (" {}: {}," * len(fit_parameter_names)).format(
            *[val for pair in zip(fit_parameter_names, beadparams) for val in pair]
        )
    )

    # Compute obj_function
    if not np.any(np.isnan(beadparams)):
        obj_function = []
        for key, data_obj in exp_dict.items():
            try:
                data_obj.update_parameters(fit_bead, fit_parameter_names, beadparams)
                obj_function.append(data_obj.objective())
            except Exception:
                logger.exception("Failed to evaluate objective function for {} of type {}.".format(key, data_obj.name))
                obj_function.append(np.inf)

        obj_total = np.nansum(obj_function)

        # Add penalty for being out of bounds for the sake of inner minimization
        for i, param in enumerate(beadparams):
            if param < bounds[i][0]:
                logger.debug("Adding penalty to {} parameter for being lower than range".format(fit_parameter_names[i]))
                obj_total += (1e3 * (param - bounds[i][0])) ** 8
            elif param > bounds[i][1]:
                logger.debug(
                    "Adding penalty to {} parameter for being higher than range".format(fit_parameter_names[i])
                )
                obj_total += (1e3 * (param - bounds[i][1])) ** 8
    else:
        logger.info("One of provided parameters, {}, is NaN".format(beadparams))
        obj_total = np.inf
        obj_function = np.nan

    if obj_total == 0.0 and np.isnan(np.sum(obj_function)):
        obj_total = np.inf

    # Write out parameters and objective functions for each dataset
    logger.info(
        "\nParameters: {}Total Obj.\nValues: {}{}\n".format(
            ("{}, " * len(fit_parameter_names)).format(*fit_parameter_names),
            ("{}, " * len(beadparams)).format(*beadparams),
            obj_total,
        )
        + "Exp. Data: {}\nObj. Values: {}".format(
            list(exp_dict.keys()),
            obj_function,
        )
    )

    return obj_total


def obj_function_form(
    data_test,
    data0,
    weights=1.0,
    method="average-squared-deviation",
    nan_number=1000,
    nan_ratio=0.1,
):
    r"""
    Sets objective functional form

    Note that if the result is np.nan, that point is removed from the list for the
    purposes of averaging.

    Parameters
    ----------
    data_test : numpy.ndarray
        Data that is being assessed. Array of data of the same length as ``data_test``
    data0 : numpy.ndarray
        Reference data for comparison
    weights : (numpy.ndarray or float), Optional, default=1.0
        Can be a float or array of data of the same length as ``data_test``. Allows
        the user to tune the importance of various data points.
    method : str, Optional, default="mean-squared-relative-error"
        Keyword used to choose the functional form. Can be:

        - average-squared-deviation:
        :math:`\sum{(\\frac{data\_test-data0}{data0})^2}/N`
        - sum-squared-deviation: :math:`\sum{(\\frac{data\_test-data0}{data0})^2}`
        - sum-squared-deviation-boltz: :math:`\sum{(\\frac{data\_test-data0}{data0})^2`
        :math:`exp(\\frac{data\_test\_min-data\_test}{|data\_test\_min|})}`
        [DOI: 10.1063/1.2181979]
        - sum-deviation-boltz: :math:`\sum{\\frac{data\_test-data0}{data0}`
        :math:`exp(\\frac{data\_test\_min-data\_test}{|data\_test\_min|})}`
        [DOI: 10.1063/1.2181979]
        - percent-absolute-average-deviation:
        :math:`\sum{(\\frac{data\_test-data0}{data0})^2}/N \\times 100`

    nan_ratio : float, Optional, default=0.1
        If more than "nan_ratio*100" percent of the calculated data failed to produce
        NaN, increase the objective value by the number of entries where data_test is
        NaN times ``nan_number``.
    nan_number : float, Optional, default=1000
        If a thermodynamic calculation produces NaN, add this quantity to the
        objective
        value. (See nan_ratio)

    Returns
    -------
    obj_value : float
        Objective value given the calculated and reference information
    """

    if np.size(data0) != np.size(data_test):
        raise ValueError(
            "Input data of length, {}, must be the same length".format(len(data_test))
            + " as reference data of length {}".format(len(data0))
        )

    if np.size(weights) > 1 and np.size(weights) != np.size(data_test):
        raise ValueError(
            "Weight for data is provided as an array of length, "
            + "{}, but must be length, {}.".format(len(weights), len(data_test))
        )

    data_tmp = np.array(
        [
            (data_test[i] - data0[i]) / data0[i]
            for i in range(len(data_test))
            if not np.isnan((data_test[i] - data0[i]) / data0[i])
        ]
    )
    if gtb.isiterable(weights):
        weight_tmp = np.array(
            [weights[i] for i in range(len(data_test)) if not np.isnan((data_test[i] - data0[i]) / data0[i])]
        )
    else:
        weight_tmp = weights

    if method == "average-squared-deviation":
        obj_value = np.mean(data_tmp**2 * weight_tmp)

    elif method == "sum-squared-deviation":
        obj_value = np.sum(data_tmp**2 * weight_tmp)

    elif method == "sum-squared-deviation-boltz":
        data_min = np.min(data_tmp)
        obj_value = np.sum(data_tmp**2 * weight_tmp * np.exp((data_min - data_tmp) / np.abs(data_min)))

    elif method == "sum-deviation-boltz":
        data_min = np.min(data_tmp)
        obj_value = np.sum(data_tmp * weight_tmp * np.exp((data_min - data_tmp) / np.abs(data_min)))

    elif method == "percent-absolute-average-deviation":
        obj_value = np.mean(np.abs(data_tmp) * weight_tmp) * 100

    if len(data_tmp) == 0:
        obj_value = np.nan

    if len(data_test) != len(data_tmp):
        tmp = 1 - len(data_tmp) / len(data_test)
        if tmp > nan_ratio:
            obj_value += (len(data_test) - len(data_tmp)) * nan_number
            logger.debug(
                "Values of NaN were removed from objective value calculation, "
                + "nan_ratio {} > {}, augment obj. value".format(tmp, nan_ratio)
            )
        else:
            logger.debug(
                "Values of NaN were removed from objective value calculation, "
                + "nan_ratio {} < {}".format(tmp, nan_ratio)
            )

    return obj_value



""" Process command line argparse and initiate logging settings.
"""

from .main import run, get_parser
from .utils.parallelization import MultiprocessingJob
import os
import logging
import logging.handlers

parser = get_parser()
args = parser.parse_args()

# Extract arguments
if args.verbose == 0:
    args.verbose = 20
elif args.verbose < 4:
    args.verbose = (4 - args.verbose) * 10
else:
    args.verbose = 10

# Logging
logger = logging.getLogger("despasito")
logger.setLevel(args.verbose)

# Set up rotating log files
if os.path.isfile(args.logFile):
    os.remove(args.logFile)
log_file_handler = logging.handlers.RotatingFileHandler(args.logFile)
log_file_handler.setFormatter(
    logging.Formatter("%(asctime)s [%(levelname)s](%(name)s:%(funcName)s:%(lineno)d): %(message)s")
)
log_file_handler.setLevel(args.verbose)
logger.addHandler(log_file_handler)

if args.console:
    # Set up logging to console
    console_handler = logging.StreamHandler()  # sys.stderr
    console_handler.setFormatter(logging.Formatter("[%(levelname)s](%(name)s): %(message)s"))
    console_handler.setLevel(args.verbose)
    logger.addHandler(console_handler)

logging.info("Input args: {}".format(args))

# Run program
if args.input:
    kwargs = {"filename": args.input}
else:
    kwargs = {}

# Update flags for optimization methods
logging.info("Use Numba JIT: {}".format(args.numba))
kwargs["numba"] = args.numba
logging.info("Use Cython: {}".format(args.cython))
kwargs["cython"] = args.cython
logging.info("Pure Python: {}".format(args.python))
kwargs["python"] = args.python

kwargs["MultiprocessingObject"] = MultiprocessingJob(ncores=args.ncores)
kwargs["path"] = args.path

run(**kwargs)



# -- coding: utf8 --

"""
    EOS object for Peng-Robinson
"""

import numpy as np

# import logging

from despasito.equations_of_state import constants
from despasito.equations_of_state.interface import EosTemplate
import despasito.utils.general_toolbox as gtb

# logger = logging.getLogger(__name__)


class EosType(EosTemplate):
    r"""
    EOS object for the Peng-Robinson EOS.

    All input and calculated parameters are defined as hidden attributes.

    Parameters
    ----------
    beads : list[str]
        List of unique component names
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters:

        - Tc: :math:`T_{C}`, Critical temperature [K]
        - Pc: :math:`P_{C}`, Critical pressure [Pa]
        - omega: :math:`\omega`, Acentric factor

    cross_library : dict, Optional, default={}
        Optional library of bead cross interaction parameters. As many or as few of the
        desired parameters may be defined for whichever group combinations are desired.

        - kij: :math:`k_{ij}`, binary interaction parameter

    Attributes
    ----------
    beads : list[str]
        List of component names
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters. See description under *Parameters*
    cross_library : dict
        Library of bead cross interaction parameters. See description under *Parameters*
    parameter_types : list[str]
        List of parameter names used Peng-Robinson EOS: ["ai", "bi", "kij", "Tc", "Pc",
        "omega"]. Used in parameter fitting.
    parameter_bound_extreme : dict
        With each parameter names as an entry representing a list with the minimum and
        maximum feasible parameter value. For the Peng-Robinson EOS:

        - ai: [0., 50.]
        - bi: [0., 1e-3]
        - kij: [-1.,1.]
        - omega: [0,1]
        - Tc: [0, 1000] [K]
        - Pc: [1, 1e+8] [Pa]

    number_of_components : int
        Number of components in mixture represented by given EOS object.
    T : float, default=numpy.nan
        Temperature value is initially defined as NaN for a placeholder until
        temperature dependent attributes are initialized by using a method of this
        class.

    """

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        self.parameter_types = ["ai", "bi", "kij", "Tc", "Pc", "omega"]
        self.parameter_bound_extreme = {
            "ai": [0.0, 50.0],
            "bi": [0.0, 1e-3],
            "kij": [-1.0, 1.0],
            "omega": [0, 1],
            "Tc": [0, 1000],
            "Pc": [1, 1e8],
        }

        # Self interaction parameters
        self.beads = kwargs["beads"]
        self.bead_library = kwargs["bead_library"]
        self.number_of_components = len(self.beads)

        self._test_kappa = [False for _ in self.beads]
        self._test_critical = [False for _ in self.beads]
        self._test_parameters = [False for _ in self.beads]
        for i, bead in enumerate(self.beads):
            if "omega" in self.bead_library[bead] and "kappa" not in self.bead_library[bead]:
                self._test_kappa[i] = True

            self._test_critical[i] = "Tc" in self.bead_library[bead] and "Pc" in self.bead_library[bead]
            self._test_parameters[i] = "ai" in self.bead_library[bead] and "bi" in self.bead_library[bead]

            if not self._test_critical[i] and not self._test_parameters[i]:
                raise ValueError("Either 'Tc' or 'Pc' was not provided for component: {}".format(bead))

        # Cross interaction parameters
        if "cross_library" in kwargs:
            self.cross_library = kwargs["cross_library"]
        else:
            self.cross_library = {}

        self.eos_dict = {
            "ai": np.zeros(self.number_of_components),
            "bi": np.zeros(self.number_of_components),
            "alpha": np.zeros(self.number_of_components),
            "aij": np.nan,
            "bij": np.nan,
            "kij": np.zeros((self.number_of_components, self.number_of_components)),
        }

        # Initialize temperature attribute
        self.T = None
        self.parameter_refresh()

    def _calc_temp_dependent_parameters(self, T):
        """
        Compute ai and alpha given temperature

        Parameters
        ----------
        T : float
            Temperature of the system [K]

        Attributes
        ----------
        ai : numpy.ndarray
            Peng-Robinson parameter a [m^6/mol^2]
        alpha : numpy.ndarray
            Peng-Robinson parameter b [m^3/mol]
        """
        if T != self.T:
            self.T = T

        for i, bead in enumerate(self.beads):
            if "kappa" in self.bead_library[bead]:
                self.eos_dict["alpha"][i] = (
                    1 + self.bead_library[bead]["kappa"] * (1 - np.sqrt(T / self.bead_library[bead]["Tc"]))
                ) ** 2
            else:
                self.eos_dict["alpha"][i] = 1.0

    def _calc_mixed_parameters(self, xi, T):
        """
        Compute mixing aij and bij given composition

        Parameters
        ----------
        xi : list[float]
            Mole fraction of each component
        T : float
            Temperature of the system [K]

        Attributes
        ----------
        aij : float
            Peng-Robinson parameter a [m^6/mol^2]
        bij : float
            Peng-Robinson parameter b [m^3/mol]
        """

        if T != self.T:
            self.T = T
            self._calc_temp_dependent_parameters(T)

        aij = 0
        index = range(len(xi))
        for i in index:
            for j in index:
                aij += (
                    xi[i]
                    * xi[j]
                    * np.sqrt(
                        self.eos_dict["ai"][i]
                        * self.eos_dict["alpha"][i]
                        * self.eos_dict["ai"][j]
                        * self.eos_dict["alpha"][j]
                    )
                    * (1.0 - self.eos_dict["kij"][i][j])
                )

        self.eos_dict["aij"] = aij
        self.eos_dict["bij"] = np.sum(xi * self.eos_dict["bi"])

    def pressure(self, rho, T, xi):
        """
        Compute pressure given system information

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : list[float]
            Mole fraction of each component

        Returns
        -------
        P : numpy.ndarray
            Array of pressure values [Pa] associated with each density and so equal
            in length
        """

        if T != self.T:
            self.T = T
            self._calc_temp_dependent_parameters(T)

        self._calc_mixed_parameters(xi, T)

        if not gtb.isiterable(rho):
            rho = np.array([rho])
        elif not isinstance(rho, np.ndarray):
            rho = np.array(rho)

        P = constants.R * self.T * rho / (1 - self.eos_dict["bij"] * rho) - rho**2 * self.eos_dict["aij"] / (
            (1 + self.eos_dict["bij"] * rho) + rho * self.eos_dict["bij"] * (1 - self.eos_dict["bij"] * rho)
        )

        return P

    def fugacity_coefficient(self, P, rho, xi, T):
        r"""
        Compute fugacity coefficient

        Parameters
        ----------
        P : float
            Pressure of the system [Pa]
        rho : float
            Molar density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : list[float]
            Mole fraction of each component

        Returns
        -------
        fugacity_coefficient : numpy.ndarray
            :math:`\phi_i`, Array of fugacity coefficient values for each component
        """

        if gtb.isiterable(T):
            if len(T) == 1:
                T = T[0]
            else:
                raise ValueError("Temperature must be given as a scalar.")
        if gtb.isiterable(rho):
            if len(rho) == 1:
                rho = rho[0]
            else:
                raise ValueError("Density must be given as a scalar.")
        if gtb.isiterable(P):
            if len(P) == 1:
                P = P[0]
            else:
                raise ValueError("Pressure must be given as a scalar.")

        if T != self.T:
            self.T = T
            self._calc_temp_dependent_parameters(T)

        self._calc_mixed_parameters(xi, T)

        tmp_RT = constants.R * T

        Z = P / (tmp_RT * rho)
        Ai = self.eos_dict["ai"] * self.eos_dict["alpha"] * P / tmp_RT**2
        Bi = self.eos_dict["bi"] * P / tmp_RT
        B = self.eos_dict["bij"] * P / tmp_RT
        A = self.eos_dict["aij"] * P / tmp_RT**2

        sqrt2 = np.sqrt(2.0)
        tmp1 = A / (2.0 * sqrt2 * B) * np.log((Z + (1 + sqrt2) * B) / (Z + (1 - sqrt2) * B))
        tmp3 = Bi * (Z - 1) / B - np.log(Z - B)
        tmp2 = np.zeros(len(xi))

        index = range(len(xi))
        for i in index:
            Aij = np.zeros(len(xi))
            for j in index:
                Aij[j] = np.sqrt(Ai[i] * Ai[j]) * (1.0 - self.eos_dict["kij"][i][j])
            tmp2[i] = Bi[i] / B - 2 * np.sum(xi * Aij) / A
        phi = np.exp(tmp1 * tmp2 + tmp3)

        return phi

    def density_max(self, xi, T, maxpack=0.9):
        """
        Estimate the maximum density based on the hard sphere packing fraction.

        Parameters
        ----------
        xi : list[float]
            Mole fraction of each component
        T : float
            Temperature of the system [K]
        maxpack : float, Optional, default=0.9
            Maximum packing fraction

        Returns
        -------
        max_density : float
            Maximum molar density [:math:`mol/m^3`]
        """

        if T != self.T:
            self.T = T
            self._calc_temp_dependent_parameters(T)

        self._calc_mixed_parameters(xi, T)

        max_density = maxpack / self.eos_dict["bij"]

        return max_density

    def update_parameter(self, param_name, bead_names, param_value):
        r"""
        Update a single parameter value during parameter fitting process.

        To refresh those parameters that are dependent on to bead_library or
        cross_library, use method "parameter refresh".

        Parameters
        ----------
        param_name : str
            Parameter to be fit. See EOS documentation for supported parameter names.
            Cross interaction parameter names should be composed of parameter name
            and the other bead type, separated by an underscore (e.g. kij_CO2).
        bead_names : list
            Bead names to be changed. For a self interaction parameter, the length will
            be 1, for a cross interaction parameter, the length will be two.
        param_value : float
            Value of parameter
        """

        if param_name in ["ai", "bi"] and self._test_critical[self.beads.index(bead_names[0])]:
            raise ValueError("Bead, {}, initialized with critical properties, not ai and bi".format(bead_names[0]))
        super().update_parameter(param_name, bead_names, param_value)

    def parameter_refresh(self):
        r"""
        To refresh dependent parameters

        Those parameters that are dependent on bead_library and cross_library
        attributes **must** be updated by running this function after all parameters
        from update_parameters method have been changed.
        """

        for i, bead in enumerate(self.beads):
            if "omega" in self.bead_library[bead] and "kappa" not in self.bead_library[bead]:
                self.bead_library[bead]["kappa"] = (
                    0.37464
                    + 1.54226 * self.bead_library[bead]["omega"]
                    - 0.26992 * self.bead_library[bead]["omega"] ** 2
                )

            if self._test_critical[i] and not self._test_parameters[i]:
                self.bead_library[bead]["ai"] = (
                    0.45723553 * (constants.R * self.bead_library[bead]["Tc"]) ** 2 / self.bead_library[bead]["Pc"]
                )
                self.bead_library[bead]["bi"] = 0.07779607 * (
                    constants.R * self.bead_library[bead]["Tc"] / self.bead_library[bead]["Pc"]
                )

            parameters = ["ai", "bi"]
            for key in parameters:
                self.eos_dict[key][i] = self.bead_library[bead][key]

            parameters = ["kij"]
            for key in parameters:
                for j, bead2 in enumerate(self.beads):
                    if (
                        bead in self.cross_library
                        and bead2 in self.cross_library[bead]
                        and key in self.cross_library[bead][bead2]
                    ):
                        tmp = self.cross_library[bead][bead2][key]
                        self.eos_dict[key][i][j] = tmp
                        self.eos_dict[key][j][i] = tmp
                    elif (
                        bead2 in self.cross_library
                        and bead in self.cross_library[bead2]
                        and key in self.cross_library[bead2][bead]
                    ):
                        tmp = self.cross_library[bead2][bead][key]
                        self.eos_dict[key][j][i] = tmp
                        self.eos_dict[key][i][j] = tmp

        if self.T is not None:
            self._calc_temp_dependent_parameters(self.T)

    def __str__(self):

        string = "Beads:" + str(self.beads) + "\n"
        string += "T:" + str(self.T) + "\n"
        return string



"""
Unit and regression test for the despasito package.
"""

# Import package, test suite, and other packages as needed
import despasito.input_output.read_input as ri
import despasito.parameter_fitting as fit
import despasito.equations_of_state
import pytest
import numpy as np

beads = ["CO2", "H2O353"]
molecular_composition = np.array([[1.0, 0.0], [0.0, 1.0]])
bead_library = {
    "H2O353": {
        "epsilon": 479.56,
        "lambdaa": 6.0,
        "lambdar": 8.0,
        "sigma": 3.0029e-1,
        "Sk": 1.0,
        "Vks": 1,
        "mass": 0.018015,
    },
    "CO2": {
        "epsilon": 353.55,
        "lambdaa": 6.66,
        "lambdar": 23.0,
        "sigma": 3.741e-1,
        "Sk": 1.0,
        "Vks": 1,
        "mass": 0.04401,
    },
}
cross_library = {"CO2": {"H2O353": {"epsilon": 432.69}}}
Eos = despasito.equations_of_state.initiate_eos(
    eos="saft.gamma_mie",
    beads=beads,
    molecular_composition=molecular_composition,
    bead_library=bead_library,
    cross_library=cross_library,
)

# Exp Data dict
Tlist = np.array([353.0])
xilist = np.array([[0.0128, 0.9872]])
yilist = np.array([[0.9896, 0.0104]])
Plist = np.array([7.08e6])

Tlist2 = np.array([353.0])
xilist2 = np.array([[0.008, 0.992]])
yilist2 = np.array([[0.9857, 0.0143]])
Plist2 = np.array([4.05e6])

exp_data_dew_pressure = {
    "Wiley": {
        "data_class_type": "TLVE",
        "eos_obj": Eos,
        "calculation_type": "dew_pressure",
        "T": Tlist2,
        "P": Plist2,
        "xi": xilist2,
        "yi": yilist2,
    }
}
exp_data_flash = {
    "Wiley": {
        "data_class_type": "flash",
        "eos_obj": Eos,
        "calculation_type": "flash",
        "T": Tlist,
        "P": Plist,
        "xi": xilist,
        "yi": yilist,
    }
}

# Optimization options
optimization_parameters = {
    "fit_bead": "CO2",
    "fit_parameter_names": ["epsilon_H2O353"],
    "epsilon_H2O353_bounds": [150.0, 600.0],
}
thermo_dict_dew = {
    "optimization_parameters": optimization_parameters,
    "exp_data": exp_data_dew_pressure,
    "parameters_guess": [432.69],
    "global_opts": {"method": "single_objective"},
}

thermo_dict_flash = {
    "optimization_parameters": optimization_parameters,
    "exp_data": exp_data_flash,
    "parameters_guess": [432.69],
    "global_opts": {"method": "single_objective"},
}


def test_dew_pressure(Eos=Eos, thermo_dict=thermo_dict_dew.copy()):

    thermo_dict = ri.process_param_fit_inputs(thermo_dict)
    output = fit.fit(**thermo_dict)

    assert output["parameters_final"][0] == pytest.approx(432.69, abs=1.0) and output[
        "objective_value"
    ] == pytest.approx(854.19, abs=1.0)


def test_flash(Eos=Eos, thermo_dict=thermo_dict_flash.copy()):

    thermo_dict = ri.process_param_fit_inputs(thermo_dict)
    output = fit.fit(**thermo_dict)

    assert output["parameters_final"][0] == pytest.approx(432.69, abs=1.0) and output[
        "objective_value"
    ] == pytest.approx(0.730, abs=0.001)



"""
Unit and regression test for the despasito package.
"""

# Import package, test suite, and other packages as needed
import despasito.input_output.read_input as ri
import despasito.parameter_fitting as fit
import despasito.equations_of_state
import pytest
import sys
import numpy as np

# EOS Object
beads = ["CH3OH"]
molecular_composition = np.array([[1.0]])
bead_library = {
    "CH3OH": {
        "epsilon": 375.01,
        "lambdaa": 6.0,
        "lambdar": 16.352,
        "sigma": 3.463e-1,
        "Sk": 1.0,
        "Vks": 2,
        "mass": 0.0310337,
    }
}
Eos = despasito.equations_of_state.initiate_eos(
    eos="saft.gamma_mie",
    beads=beads,
    molecular_composition=molecular_composition,
    bead_library=bead_library,
)

beads = ["CH3", "CH2"]
molecular_composition = np.array([[2.0, 4.0], [2.0, 10.0]])
bead_library = {
    "CH3": {
        "epsilon": 256.7662,
        "lambdaa": 6.0,
        "lambdar": 15.04982,
        "sigma": 4.077257e-1,
        "Sk": 0.5725512,
        "Vks": 1,
        "mass": 0.01503502,
    },
    "CH2": {
        "epsilon": 473.3893,
        "lambdaa": 6.0,
        "lambdar": 19.87107,
        "sigma": 4.880081e-1,
        "Sk": 0.2293202,
        "Vks": 1,
        "mass": 0.01402708,
    },
}
cross_library = {"CH3": {"CH2": {"epsilon": 350.77}}}
Eos_mix = despasito.equations_of_state.initiate_eos(
    eos="saft.gamma_mie",
    beads=beads,
    molecular_composition=molecular_composition,
    bead_library=bead_library,
    cross_library=cross_library,
)


# Exp Data dict
exp_data_sat = {
    "Wiley": {
        "data_class_type": "saturation_properties",
        "eos_obj": Eos,
        "calculation_type": "saturation_properties",
        "T": np.array([200.0]),
        "Psat": np.array([6.1000e01]),
        "rhol": np.array([27474.40699]),
        "rhov": np.array([3.12109900e-03]),
    },
    "Gibbard": {
        "data_class_type": "saturation_properties",
        "eos_obj": Eos,
        "calculation_type": "saturation_properties",
        "T": np.array([288.1506, 303.1427]),
        "Psat": np.array([9884.4, 21874.3]),
    },
}

exp_data_sol = {
    "Wiley": {
        "data_class_type": "solubility_parameter",
        "eos_obj": Eos,
        "calculation_type": "solubility_parameter",
        "T": np.array([288.1506]),
        "P": np.array([11152.285]),
        "rhol": np.array([24098.4771]),
        "delta": np.array([29161.4886]),
    }
}

exp_data_density = {
    "dx.doi.org/10.1021/acs.jced.0c00943": {
        "data_class_type": "liquid_density",
        "eos_obj": Eos_mix,
        "calculation_type": "liquid_properties",
        "T": np.array([293.15]),
        "xi": np.array([[0.1999, 0.8001], [0.7998, 0.2002]]),
        "P": np.array([5e6]),
        "rhol": np.array([4844.03412, 6863.260764]),
    }
}

# Optimization options
optimization_parameters = {
    "fit_bead": "CH3OH",
    "fit_parameter_names": ["epsilon"],
    "epsilon_bounds": [300.0, 400.0],
}
thermo_dict0 = {
    "optimization_parameters": optimization_parameters,
    "parameters_guess": [384.0],
    "global_opts": {"method": "single_objective"},
}

optimization_parameters_mix = {
    "fit_bead": "CH3",
    "fit_parameter_names": ["epsilon_CH2"],
    "epsilon_bounds": [300.0, 400.0],
}
thermo_dict_mix = {
    "optimization_parameters": optimization_parameters_mix,
    "parameters_guess": [350],
    "global_opts": {"method": "single_objective"},
}


def test_fit_import():
    #    """Sample test, will always pass so long as import statement worked"""
    assert "despasito.parameter_fitting" in sys.modules


thermo_dict0["exp_data"] = exp_data_sol


def test_solubility_so(Eos=Eos, thermo_dict=thermo_dict0.copy()):

    thermo_dict = ri.process_param_fit_inputs(thermo_dict)
    output = fit.fit(**thermo_dict)

    assert output["parameters_final"][0] == pytest.approx(375.01, abs=1.0) and output["objective_value"] < 1.1


thermo_dict_mix["exp_data"] = exp_data_density


def test_density_so(Eos=Eos_mix, thermo_dict=thermo_dict_mix.copy()):

    thermo_dict = ri.process_param_fit_inputs(thermo_dict)
    output = fit.fit(**thermo_dict)

    assert output["parameters_final"][0] == pytest.approx(350.0, abs=1.0) and output["objective_value"] < 1.5


thermo_dict0["exp_data"] = exp_data_sat
thermo_dict0["global_opts"] = {"method": "single_objective"}


def test_saturation_de(Eos=Eos, thermo_dict=thermo_dict0.copy()):

    thermo_dict["density_opts"] = {"pressure_min": 10}
    thermo_dict = ri.process_param_fit_inputs(thermo_dict)
    output = fit.fit(**thermo_dict)

    assert output["parameters_final"][0] == pytest.approx(375.01, abs=1.0) and output[
        "objective_value"
    ] == pytest.approx(5.7658, abs=0.01)



# -- coding: utf8 --
r"""
    EOS object for SAFT-:math:`\gamma`-SW

    Equations referenced in this code are from Lymperiadis, A. et. al, J. Chem. Phys.
    127, 234903 (2007)
"""

import numpy as np
import logging

import despasito.equations_of_state.eos_toolbox as tb
from despasito.equations_of_state import constants
import despasito.equations_of_state.saft.saft_toolbox as stb
from despasito.equations_of_state.saft import Aassoc

logger = logging.getLogger(__name__)

ckl_coef = np.array(
    [
        [2.25855, -1.50349, 0.249434],
        [-0.669270, 1.40049, -0.827739],
        [10.1576, -15.0427, 5.30827],
    ]
)


class SaftType:
    r"""
    Object of SAFT-𝛾-SW (for square well potential)

    Parameters
    ----------
    beads : list[str]
        List of unique bead names used among components
    molecular_composition : numpy.ndarray
        :math:`\nu_{i,k}/k_B`. Array containing the number of components by the number
        of bead types. Defines the number of each type of group in each component.
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters:

        - epsilon: :math:`\epsilon_{k,k}/k_B`, Energy well depth scaled by Boltzmann
        constant
        - sigma: :math:`\sigma_{k,k}`, Size parameter, contact distance [nm]
        - lambda: :math:`\lambda_{k,k}`, Range of the attractive interaction of well
        depth, epsilon
        - Sk: Optional, default=1, Shape factor, reflects the proportion which a given
        segment contributes to the total free energy
        - Vks: Optional, default=1, Number of segments in this molecular group

    cross_library : dict, Optional, default={}
        Optional library of bead cross interaction parameters. As many or as few of
        the desired parameters may be defined for whichever group combinations are
        desired. If this matrix isn't provided, the SAFT ``combining_rules`` are used.

        - epsilon: :math:`\epsilon_{k,l}/k_B`, Energy parameter, well depth, scaled
        by Boltzmann Constant
        - sigma: :math:`\sigma_{k,k}`, Size parameter, contact distance [nm]
        - lambda: :math:`\lambda_{k,l}`, Range of the attractive interaction of well
        depth, epsilon

    num_rings : list
        Number of rings in each molecule. This will impact the chain contribution to
        the Helmholtz energy.

    Attributes
    ----------
    beads : list[str]
        List of unique bead names used among components
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters. See entry in **Parameters** section.
    cross_library : dict
        Library of bead cross interaction parameters. As many or as few of the desired
        parameters may be defined for whichever group combinations are desired. See
        entry in **Parameters** section.
    Aideal_method : str
        "Abroglie" the default functional form of the ideal gas contribution of the
        Helmholtz energy
    residual_helmholtz_contributions : list[str]
        List of methods from the specified saft_source representing contributions to
        the Helmholtz energy that are functions of density, temperature, and
        composition. For this variant, [``Amonomer``, ``Achain``]
    parameter_types : list[str]
        This list of parameter names, "epsilon", "lambda", "sigma", and/or "Sk" as well
        as parameters for the specific SAFT variant.
    parameter_bound_extreme : dict
        With each parameter name as an entry representing a list with the minimum and
        maximum feasible parameter value.

        - epsilon: [10.,1000.]
        - lambda: [1.0,10.]
        - sigma: [0.1,10.0]
        - Sk: [0.1,1.0]

    combining_rules : dict
        Contains functional form and additional information for calculating cross
        interaction parameters that are not found in ``cross_library``. Function must
        be one of those contained in
        :mod:`~despasito.equations_of_state.combining_rule_types`. The default values
        are:

        - sigma: {"function": "mean"}
        - lambda: {"function": "weighted_mean","weighting_parameters": ["sigma"]}
        - epsilon: {"function": "square_well_berthelot","weighting_parameters":
        ["sigma", "lambda"]}

    eos_dict : dict
        Dictionary of parameters and specific settings

        - molecular_composition (numpy.ndarray) - :math:`\nu_{i,k}/k_B`. Array
        containing the number of components by the number of bead types. Defines the
        number of each type of group in each component.
        - num_rings (list) - Number of rings in each molecule. This will impact the
        chain contribution to the Helmholtz energy.
        - Sk (numpy.ndarray) - Shape factor, reflects the proportion which a given
        segment contributes to the total free energy. Length of ``beads`` array.
        - Vks (numpy.ndarray) - Number of segments in this molecular group. Length of
        ``beads`` array.
        - epsilon_kl (numpy.ndarray) - Matrix of well depths for groups (k,l)
        - sigma_kl (numpy.ndarray) - Matrix of bead diameters (k,l)
        - lambda_kl (numpy.ndarray) - Matrix of range of potential well depth (k,l)
        - Cmol2seg (float) - Conversion factor from from molecular number density,
        :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`.
        - epsilon_ij (numpy.ndarray) - Matrix of average molecular well depths (k,l)
        - sigma_ij (numpy.ndarray) - Matrix of average molecular diameter (k,l)
        - lambda_ij (numpy.ndarray) - Matrix of average molecular range of potential
        well depth (k,l)
        - xskl (numpy.ndarray) - Matrix of mole fractions of bead (i.e. segment or
        group) k multiplied by that of bead l
        - alphakl (np.array) - van der Waals attractive parameter for square-well
        segments, equal to :math:`\alpha_{k,l}/k_B`.

    ncomp : int
        Number of components in the system
    nbeads : int
        Number of beads in system that are shared among components
    xi : numpy.ndarray
        Mole fraction of each molecule in mixture. Default initialization is np.nan
    """

    def __init__(self, **kwargs):

        if "method_stat" in kwargs:
            self.method_stat = kwargs["method_stat"]
            del kwargs["method_stat"]
            logger.info("This EOS doesn't use compiled modules for Amonomer and Achain")
        else:
            self.method_stat = None

        self.Aideal_method = "Abroglie"
        self.residual_helmholtz_contributions = ["Amonomer", "Achain"]
        self.parameter_types = ["epsilon", "lambda", "sigma", "Sk"]
        self._parameter_defaults = {
            "epsilon": None,
            "lambda": None,
            "sigma": None,
            "Sk": 1.0,
            "Vks": 1.0,
        }
        self.parameter_bound_extreme = {
            "epsilon": [10.0, 1000.0],
            "lambda": [1.0, 10.0],
            "sigma": [0.1, 10.0],
            "Sk": [0.1, 1.0],
        }
        self.combining_rules = {
            "sigma": {"function": "mean"},
            "lambda": {"function": "weighted_mean", "weighting_parameters": ["sigma"]},
            "epsilon": {
                "function": "square_well_berthelot",
                "weighting_parameters": ["sigma", "lambda"],
            },
        }  # Note in this EOS object, the combining rules for the group parameters
        # are also used for their corresponding molecular averaged parameters.

        if not hasattr(self, "eos_dict"):
            self.eos_dict = {}

        needed_attributes = ["molecular_composition", "beads", "bead_library"]
        for key in needed_attributes:
            if key not in kwargs:
                raise ValueError("The one of the following inputs is missing: {}".format(", ".join(needed_attributes)))
            elif key == "molecular_composition":
                if "molecular_composition" not in self.eos_dict:
                    self.eos_dict[key] = kwargs[key]
            elif not hasattr(self, key):
                setattr(self, key, kwargs[key])

        self.bead_library = tb.check_bead_parameters(self.bead_library, self._parameter_defaults)

        if "cross_library" not in kwargs:
            self.cross_library = {}
        else:
            self.cross_library = kwargs["cross_library"]

        if "Vks" not in self.eos_dict:
            self.eos_dict["Vks"] = tb.extract_property("Vks", self.bead_library, self.beads, default=1.0)
        if "Sk" not in self.eos_dict:
            self.eos_dict["Sk"] = tb.extract_property("Sk", self.bead_library, self.beads, default=1.0)

        # Initialize component attribute
        if not hasattr(self, "xi"):
            self.xi = np.nan
        if not hasattr(self, "nbeads") or not hasattr(self, "ncomp"):
            self.ncomp, self.nbeads = np.shape(self.eos_dict["molecular_composition"])

        # Initiate cross interaction terms
        output = tb.cross_interaction_from_dict(
            self.beads,
            self.bead_library,
            self.combining_rules,
            cross_library=self.cross_library,
        )
        self.eos_dict["sigma_kl"] = output["sigma"]
        self.eos_dict["epsilon_kl"] = output["epsilon"]
        self.eos_dict["lambda_kl"] = output["lambda"]

        if "num_rings" in kwargs:
            self.eos_dict["num_rings"] = kwargs["num_rings"]
            logger.info("Accepted component ring structure: {}".format(kwargs["num_rings"]))
        else:
            self.eos_dict["num_rings"] = np.zeros(len(self.eos_dict["molecular_composition"]))

        # Initiate average interaction terms
        self.calc_component_averaged_properties()
        self.alphakl = (
            2.0
            * np.pi
            / 3.0
            * self.eos_dict["epsilon_kl"]
            * self.eos_dict["sigma_kl"] ** 3
            * (self.eos_dict["lambda_kl"] ** 3 - 1.0)
        )

    def calc_component_averaged_properties(self):
        r"""
        Calculate component averaged properties specific to SAFT-𝛾-SW

        Attributes
        ----------
        eos_dict : dict
            Dictionary of outputs, the following possibilities are calculated if all
            relevant beads have those properties.

            - epsilon_ij (numpy.ndarray) - Matrix of average molecular well
            depths (k,l)
            - sigma_ij (numpy.ndarray) - Matrix of average molecular diameter (k,l)
            - lambda_ij (numpy.ndarray) - Matrix of average molecular range of
            potential well depth (k,l)

        """

        ncomp, nbeads = np.shape(self.eos_dict["molecular_composition"])
        zki = np.zeros((ncomp, nbeads), float)
        zkinorm = np.zeros(ncomp, float)

        epsilonii = np.zeros(ncomp, float)
        sigmaii = np.zeros(ncomp, float)
        lambdaii = np.zeros(ncomp, float)

        # compute zki
        for i in range(ncomp):
            for k in range(nbeads):
                zki[i, k] = (
                    self.eos_dict["molecular_composition"][i, k] * self.eos_dict["Vks"][k] * self.eos_dict["Sk"][k]
                )
                zkinorm[i] += zki[i, k]

        for i in range(ncomp):
            for k in range(nbeads):
                zki[i, k] = zki[i, k] / zkinorm[i]

        for i in range(ncomp):
            for k in range(nbeads):
                sigmaii[i] += zki[i, k] * self.eos_dict["sigma_kl"][k, k] ** 3
                for l in range(nbeads):

                    epsilonii[i] += zki[i, k] * zki[i, l] * self.eos_dict["epsilon_kl"][k, l]
                    lambdaii[i] += zki[i, k] * zki[i, l] * self.eos_dict["lambda_kl"][k, l]
            sigmaii[i] = sigmaii[i] ** (1.0 / 3.0)

        input_dict = {"sigma": sigmaii, "lambda": lambdaii, "epsilon": epsilonii}
        dummy_dict, dummy_labels = tb.construct_dummy_bead_library(input_dict)
        output_dict = tb.cross_interaction_from_dict(dummy_labels, dummy_dict, self.combining_rules)
        self.eos_dict["sigma_ij"] = output_dict["sigma"]
        self.eos_dict["lambda_ij"] = output_dict["lambda"]
        self.eos_dict["epsilon_ij"] = output_dict["epsilon"]

    def reduced_density(self, rho, xi):
        r"""
        Reduced density matrix where the segment number density is reduced by powers
        of the size parameter, sigma.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Returns
        -------
        zeta : numpy.ndarray
            Reduced density matrix of length 4 of varying degrees of dependence on
            sigma. Units: [molecules/nm^3, molecules/nm^2, molecules/nm, molecules]
        """

        self._check_density(rho)
        self._check_composition_dependent_parameters(xi)

        rho2 = rho * constants.molecule_per_nm3 * self.eos_dict["Cmol2seg"]

        reduced_density = np.zeros((np.size(rho), 4))
        for m in range(4):
            reduced_density[:, m] = rho2 * (
                np.sum(np.sqrt(np.diag(self.eos_dict["xskl"])) * (np.diag(self.eos_dict["sigma_kl"]) ** m))
                * (np.pi / 6.0)
            )

        return reduced_density

    def effective_packing_fraction(self, rho, xi, zetax=None, mode="normal"):
        r"""
        Effective packing fraction for SAFT-gamma with a square-wave potential

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        zetax : numpy.ndarray, Optional, default=None
            Matrix of hypothetical packing fraction based on hard sphere diameter
            for groups (k,l)
        mode : str, Optional, default="normal"
            This indicates whether group or effective component parameters are used.
            Options include: "normal" and "effective"

        Returns
        -------
        zeta_eff : numpy.ndarray
            Effective packing fraction (len(rho), Nbeads, Nbeads)
        """

        self._check_density(rho)
        self._check_composition_dependent_parameters(xi)

        if mode == "normal":
            lambdakl = self.eos_dict["lambda_kl"]
        elif mode == "effective":
            lambdakl = self.eos_dict["lambda_ij"]
        lx = len(lambdakl)  # lx is nbeads for normal and ncomp for effective

        if zetax is None:
            zetax = self.reduced_density(rho, xi)[:, 3]

        zetax_pow = np.zeros((np.size(rho), 3))
        zetax_pow[:, 0] = zetax
        for i in range(1, 3):
            zetax_pow[:, i] = zetax_pow[:, i - 1] * zetax_pow[:, 0]

        zetakl = np.zeros((np.size(rho), lx, lx))
        for k in range(lx):
            for l in range(lx):
                if lambdakl[k, l] != 0.0:
                    cikl = np.dot(
                        ckl_coef,
                        np.array(
                            (1.0, lambdakl[k, l], lambdakl[k, l] ** 2),
                            dtype=ckl_coef.dtype,
                        ),
                    )
                    zetakl[:, k, l] = np.dot(zetax_pow, cikl)

        return zetakl

    def _dzetaeff_dzetax(self, rho, xi, zetax=None, mode="normal"):
        r"""
        Derivative of effective packing fraction with respect to the reduced density.
        Eq. 33

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        zetax : numpy.ndarray, Optional, default=None
            Matrix of hypothetical packing fraction based on hard sphere diameter for
            groups (k,l)
        mode : str, Optional, default="normal"
            This indicates whether group or effective component parameters are used.
            Options include: "normal" and "effective"

        Returns
        -------
        dzetakl : numpy.ndarray
            Derivative of effective packing fraction (len(rho), Nbeads, Nbeads) with
            respect to the reduced density
        """

        self._check_density(rho)
        self._check_composition_dependent_parameters(xi)

        if mode == "normal":
            lambdakl = self.eos_dict["lambda_kl"]
        elif mode == "effective":
            lambdakl = self.eos_dict["lambda_ij"]
        lx = len(lambdakl)  # lx is nbeads for normal and ncomp for effective

        if zetax is None:
            zetax = self.reduced_density(rho, xi)[:, 3]

        zetax_pow = np.transpose(np.array([np.ones(len(rho)), 2 * zetax, 3 * zetax**2]))

        # check if you have more than 1 bead types
        dzetakl = np.zeros((np.size(rho), lx, lx))
        for k in range(lx):
            for l in range(lx):
                if lambdakl[k, l] != 0.0:
                    cikl = np.dot(
                        ckl_coef,
                        np.array(
                            (1.0, lambdakl[k, l], lambdakl[k, l] ** 2),
                            dtype=ckl_coef.dtype,
                        ),
                    )
                    dzetakl[:, k, l] = np.dot(zetax_pow, cikl)

        return dzetakl

    def Ahard_sphere(self, rho, T, xi):
        r"""
        Outputs hard sphere approximation of Helmholtz free energy,
        :math:`A^{HS}/Nk_{b}T`.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Returns
        -------
        Ahard_sphere : numpy.ndarray
            Helmholtz energy of monomers for each density given.
        """

        rho = self._check_density(rho)
        self._check_composition_dependent_parameters(xi)

        zeta = self.reduced_density(rho, xi)

        tmp = 6.0 / (np.pi * rho * constants.molecule_per_nm3)
        tmp1 = np.log1p(-zeta[:, 3]) * (zeta[:, 2] ** 3 / (zeta[:, 3] ** 2) - zeta[:, 0])
        tmp2 = 3.0 * zeta[:, 2] / (1 - zeta[:, 3]) * zeta[:, 1]
        tmp3 = zeta[:, 2] ** 3 / (zeta[:, 3] * ((1.0 - zeta[:, 3]) ** 2))
        AHS = tmp * (tmp1 + tmp2 + tmp3)

        return AHS

    def Afirst_order(self, rho, T, xi, zetax=None):
        r"""
        Outputs :math:`A^{1st order}/Nk_{b}T`. This is the first order term in the
        high-temperature perturbation expansion

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        zetax : numpy.ndarray, Optional, default=None
            Matrix of hypothetical packing fraction based on hard sphere diameter
            for groups (k,l)

        Returns
        -------
        Afirst_order : numpy.ndarray
            Helmholtz energy of monomers for each density given.
        """

        rho = self._check_density(rho)
        self._check_composition_dependent_parameters(xi)

        if zetax is None:
            zetax = self.reduced_density(rho, xi)[:, 3]

        g0HS = self.calc_g0HS(rho, xi, zetax=zetax)
        a1kl_tmp = np.tensordot(rho * constants.molecule_per_nm3, self.eos_dict["xskl"] * self.alphakl, 0)
        A1 = -(self.eos_dict["Cmol2seg"] ** 2 / T) * np.sum(a1kl_tmp * g0HS, axis=(1, 2))  # Units of K

        return A1

    def Asecond_order(self, rho, T, xi, zetax=None, KHS=None):
        r"""
        Outputs :math:`A^{2nd order}/Nk_{b}T`. This is the second order term in the
        high-temperature perturbation expansion

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        zetax : numpy.ndarray, Optional, default=None
            Matrix of hypothetical packing fraction based on hard sphere diameter
            for groups (k,l)
        KHS : numpy.ndarray, Optional, default=None
            (length of densities) isothermal compressibility of system with packing
            fraction zetax

        Returns
        -------
        Asecond_order : numpy.ndarray
            Helmholtz energy of monomers for each density given.
        """

        rho = self._check_density(rho)
        self._check_composition_dependent_parameters(xi)

        if zetax is None:
            zetax = self.reduced_density(rho, xi)[:, 3]
        # Note that zetax = zeta3

        if KHS is None:
            KHS = stb.calc_KHS(zetax)

        dzetakl = self._dzetaeff_dzetax(rho, xi, zetax=zetax)
        zeta_eff = self.effective_packing_fraction(rho, xi, zetax=zetax)
        g0HS = self.calc_g0HS(rho, xi, zetax=zetax)

        rho2 = self.eos_dict["Cmol2seg"] * rho * constants.molecule_per_nm3

        tmp1 = KHS * rho2 / 2.0
        tmp2 = self.eos_dict["epsilon_kl"] * self.alphakl * self.eos_dict["xskl"]
        a2kl_tmp = np.tensordot(tmp1, tmp2, 0)

        a2 = a2kl_tmp * (g0HS + zetax[:, np.newaxis, np.newaxis] * dzetakl * (2.5 - zeta_eff) / (1 - zeta_eff) ** 4)

        # Lymperiadis 2007 has a disconnect where Eq. 24 != Eq. 30, as Eq. 24 is
        # missing a minus sign. (Same in Lymperiadis 2008 for Eq. 32 and Eq. 38)
        A2 = -(self.eos_dict["Cmol2seg"] / (T**2)) * np.sum(a2, axis=(1, 2))

        return A2

    def Amonomer(self, rho, T, xi):
        r"""
        Outputs the monomer contribution of the Helmholtz energy
        :math:`A^{mono.}/Nk_{b}T`.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Returns
        -------
        Amonomer : numpy.ndarray
            Helmholtz energy of monomers for each density given.
        """

        if np.all(rho > self.density_max(xi, T)):
            raise ValueError(
                "Density values should not all be greater than {}, or calc_Amono will"
                "fail in log calculation.".format(self.density_max(xi, T))
            )

        rho = self._check_density(rho)
        self._check_composition_dependent_parameters(xi)

        zetax = self.reduced_density(rho, xi)[:, 3]

        Amonomer = (
            self.Ahard_sphere(rho, T, xi)
            + self.Afirst_order(rho, T, xi, zetax=zetax)
            + self.Asecond_order(rho, T, xi, zetax=zetax)
        )

        return Amonomer

    def calc_g0HS(self, rho, xi, zetax=None, mode="normal"):
        r"""
        The contact value of the pair correlation function of a hypothetical pure fluid
        of diameter sigmax evaluated at an effective packing fraction, zeta_eff.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        zetax : numpy.ndarray, Optional, default=None
            Matrix of hypothetical packing fraction based on hard sphere diameter for
            groups (k,l)
        mode : str, Optional, default="normal"
            This indicates whether group or effective component parameters are used.
            Options include: "normal" and "effective", where normal used bead
            interaction matrices, and effective uses component averaged parameters.

        Returns
        -------
        g0HS : numpy.ndarray
            The contact value of the pair correlation function of a hypothetical pure
            fluid
        """

        rho = self._check_density(rho)
        self._check_composition_dependent_parameters(xi)

        if zetax is None:
            zetax = self.reduced_density(rho, xi)[:, 3]

        zeta_eff = self.effective_packing_fraction(rho, xi, mode=mode, zetax=zetax)

        g0HS = (1.0 - zeta_eff / 2.0) / (1.0 - zeta_eff) ** 3

        return g0HS

    def calc_gHS(self, rho, xi):
        r"""
        Hypothetical pair correlation function of a hypothetical pure fluid.

        This fluid is of diameter sigmax evaluated at contact and effective packing
        fraction zeta_eff.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Returns
        -------
        gHS : numpy.ndarray
            Hypothetical pair correlation function of a hypothetical pure fluid of
            diameter sigmax evaluated at contact and effective packing fraction
            zeta_eff.
        """

        rho = self._check_density(rho)
        self._check_composition_dependent_parameters(xi)

        zetam = self.reduced_density(rho, xi)

        tmp1 = 1.0 / (1.0 - zetam[:, 3])
        tmp2 = zetam[:, 2] / (1.0 - zetam[:, 3]) ** 2
        tmp3 = zetam[:, 2] ** 2 / (1.0 - zetam[:, 3]) ** 3

        gHS = np.zeros((np.size(rho), self.ncomp, self.ncomp))
        for i in range(self.ncomp):
            for j in range(self.ncomp):
                tmp = (
                    self.eos_dict["sigma_ij"][i, i]
                    * self.eos_dict["sigma_ij"][j, j]
                    / (self.eos_dict["sigma_ij"][i, i] + self.eos_dict["sigma_ij"][j, j])
                )
                gHS[:, i, j] = tmp1 + 3 * tmp * tmp2 + 2 * tmp**2 * tmp3

        return gHS

    def calc_gSW(self, rho, T, xi, zetax=None):
        r"""
        Calculate the square-well pair correlation function at the effective contact
        distance and the actual packing fraction of the mixture.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        zetax : numpy.ndarray, Optional, default=None
            Matrix of hypothetical packing fraction based on hard sphere diameter for
            groups (k,l)

        Returns
        -------
        gSW : numpy.ndarray
            Square-well pair correlation function at the effective contact distance and
            the actual packing fraction of the mixture.
        """

        rho = self._check_density(rho)
        self._check_composition_dependent_parameters(xi)

        if zetax is None:
            zetax = self.reduced_density(rho, xi)[:, 3]

        g0HS = self.calc_g0HS(rho, xi, zetax=zetax, mode="effective")
        gHS = self.calc_gHS(rho, xi)
        zeta_eff = self.effective_packing_fraction(rho, xi, mode="effective", zetax=zetax)
        dg0HSdzetaeff = (2.5 - zeta_eff) / (1.0 - zeta_eff) ** 4

        ncomp = len(xi)
        dckl_coef = np.array([[-1.50349, 0.249434], [1.40049, -0.827739], [-15.0427, 5.30827]])
        zetax_pow = np.transpose(np.array([zetax, zetax**2, zetax**3]))
        dzetaijdlambda = np.zeros((np.size(rho), ncomp, ncomp))
        for i in range(ncomp):
            for j in range(ncomp):
                cikl = np.dot(dckl_coef, np.array([1.0, (2 * self.eos_dict["lambda_ij"][i, j])]))
                dzetaijdlambda[:, i, j] = np.dot(zetax_pow, cikl)

        dzetaijdzetax = self._dzetaeff_dzetax(rho, xi, zetax=zetax, mode="effective")
        dzetaeff = (
            self.eos_dict["lambda_ij"][np.newaxis, :, :] / 3.0 * dzetaijdlambda
            - zetax[:, np.newaxis, np.newaxis] * dzetaijdzetax
        )

        gSW = gHS + self.eos_dict["epsilon_ij"][np.newaxis, :, :] / T * (
            g0HS + (self.eos_dict["lambda_ij"][np.newaxis, :, :] ** 3 - 1.0) * dg0HSdzetaeff * dzetaeff
        )

        return gSW

    def Achain(self, rho, T, xi):
        r"""
        Outputs chain contribution to the Helmholtz energy :math:`A^{chain}/Nk_{b}T`.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Returns
        -------
        Achain : numpy.ndarray
            Helmholtz energy of monomers for each density given.
        """

        rho = self._check_density(rho)

        gii = self.calc_gSW(rho, T, xi)

        Achain = 0.0
        for i in range(self.ncomp):
            beadsum = -1.0 + self.eos_dict["num_rings"][i]
            for k in range(self.nbeads):
                beadsum += (
                    self.eos_dict["molecular_composition"][i, k] * self.eos_dict["Vks"][k] * self.eos_dict["Sk"][k]
                )
            Achain -= xi[i] * beadsum * np.log(gii[:, i, i])

        if np.any(np.isnan(Achain)):
            logger.error("Some Helmholtz values are NaN, check energy parameters.")

        return Achain

    def density_max(self, xi, T, maxpack=0.65):
        """
        Estimate the maximum density based on the hard sphere packing fraction.

        Parameters
        ----------
        xi : list[float]
            Mole fraction of each component
        T : float
            Temperature of the system [K]
        maxpack : float, Optional, default=0.65
            Maximum packing fraction

        Returns
        -------
        max_density : float
            Maximum molar density [:math:`mol/m^3`]
        """

        self._check_composition_dependent_parameters(xi)

        # estimate the maximum density based on the hard sphere packing fraction
        # etax, assuming a maximum packing fraction specified by maxpack
        max_density = (
            maxpack
            * 6.0
            / (self.eos_dict["Cmol2seg"] * np.pi * np.sum(self.eos_dict["xskl"] * (self.eos_dict["sigma_kl"] ** 3)))
            / constants.molecule_per_nm3
        )

        return max_density

    def calc_gr_assoc(self, rho, T, xi, Ktype="ijklab"):
        r"""
        Reference fluid pair correlation function used in calculating association sites

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        Ktype : str, Optional, default='ijklab'
            Indicates which radial distribution function to return. The only option is
            'ijklab': The bonding volume was calculated from self.calc_Kijklab, return
            gHS_dij)

        Returns
        -------
        gr : numpy.ndarray
            A temperature-density polynomial correlation of the association integral
            for a Lennard-Jones monomer. This matrix is (len(rho) x Ncomp x Ncomp)
        """

        rho = self._check_density(rho)
        gSW = self.calc_gSW(rho, T, xi)

        return gSW

    def calc_Kijklab(self, T, rc_klab, rd_klab=None, reduction_ratio=0.25):
        r"""
        Calculation of association site bonding volume, dependent on molecule in
        addition to group

        Lymperiadis Fluid Phase Equilibria 274 (2008) 85–104

        Parameters
        ----------
        T : float
            Temperature of the system [K], Note used in this version of saft, but
            included to allow saft.py to be general
        rc_klab : numpy.ndarray
            This matrix of cutoff distances for association sites for each site type
            in each group type
        rd_klab : numpy.ndarray, Optional, default=None
            Position of association site in each group (nbead, nbead, nsite, nsite)
        reduction_ratio : float, Optional, default=0.25
            Reduced distance of the sites from the center of the sphere of interaction.
            This value is used when site position, rd_klab is None

        Returns
        -------
        Kijklab : numpy.ndarray
            Matrix of binding volumes
        """

        dij_bar = np.zeros((self.ncomp, self.ncomp))
        for i in range(self.ncomp):
            for j in range(self.ncomp):
                dij_bar[i, j] = np.mean([self.eos_dict["sigma_ij"][i], self.eos_dict["sigma_ij"][j]])

        Kijklab = Aassoc.calc_bonding_volume(rc_klab, dij_bar, rd_klab=rd_klab, reduction_ratio=reduction_ratio)

        return Kijklab

    def parameter_refresh(self, bead_library, cross_library):
        r"""
        To refresh dependent parameters

        Those parameters that are dependent on bead_library and cross_library
        attributes **must** be updated by running this function after all parameters
        from update_parameters method have been changed.

        Attributes
        ----------
        alpha : np.array
            van der Waals attractive parameter for square-well segments, equal to
            :math:`\alpha_{k,l}/k_B`.
        eos_dict : dict
            The following entries are updated:

            - epsilon_kl (numpy.ndarray) - Matrix of well depths for groups (k,l)
            - sigma_kl (numpy.ndarray) - Matrix of bead diameters (k,l)
            - lambda_kl (numpy.ndarray) - Matrix of range of potential well
            depth (k,l)
            - xskl (numpy.ndarray) - Matrix of mole fractions of bead (i.e. segment or
            group) k multiplied by that of bead l
            - Cmol2seg (float) - Conversion factor from from molecular number density,
            :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`.

        """

        self.bead_library.update(bead_library)
        self.cross_library.update(cross_library)

        self.eos_dict["Sk"] = tb.extract_property("Sk", self.bead_library, self.beads, default=1.0)

        # Update Non bonded matrices
        output = tb.cross_interaction_from_dict(
            self.beads,
            self.bead_library,
            self.combining_rules,
            cross_library=self.cross_library,
        )
        self.eos_dict["sigma_kl"] = output["sigma"]
        self.eos_dict["epsilon_kl"] = output["epsilon"]
        self.eos_dict["lambda_kl"] = output["lambda"]
        self.calc_component_averaged_properties()

        if not np.any(np.isnan(self.xi)):
            self.eos_dict["Cmol2seg"], self.eos_dict["xskl"] = stb.calc_composition_dependent_variables(
                self.xi,
                self.eos_dict["molecular_composition"],
                self.bead_library,
                self.beads,
            )
        self.alphakl = (
            2.0
            * np.pi
            / 3.0
            * self.eos_dict["epsilon_kl"]
            * self.eos_dict["sigma_kl"] ** 3
            * (self.eos_dict["lambda_kl"] ** 3 - 1.0)
        )

    def _check_density(self, rho):
        r"""
        This function checks that the density array is in the correct format for
        further calculations.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]

        Returns
        -------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        """

        if np.isscalar(rho):
            rho = np.array([rho])
        elif not isinstance(rho, np.ndarray):
            rho = np.array(rho)
        if len(np.shape(rho)) == 2:
            rho = rho[0]

        if any(np.isnan(rho)):
            raise ValueError("NaN was given as a value of density, rho")
        elif rho.size == 0:
            raise ValueError("No value of density was given")
        elif any(rho < 0.0):
            raise ValueError("Density values cannot be negative.")

        return rho

    def _check_composition_dependent_parameters(self, xi):
        r"""
        This function updates composition dependent variables

        Parameters
        ----------
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Attributes
        ----------
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        eos_dict : dict
            The following entries are updated:

            - xskl (numpy.ndarray) - Matrix of mole fractions of bead (i.e. segment or
            group) k multiplied by that of bead l
            - Cmol2seg (float) - Conversion factor from from molecular number density,
            :math:`\rho`, to segment (i.e. group) number density, :math:`\rho_S`.

        """
        xi = np.array(xi)
        if not np.all(self.xi == xi):
            self.eos_dict["Cmol2seg"], self.eos_dict["xskl"] = stb.calc_composition_dependent_variables(
                xi,
                self.eos_dict["molecular_composition"],
                self.bead_library,
                self.beads,
            )
            self.xi = xi

    def __str__(self):

        string = "Beads: {}".format(self.beads)
        return string



# -- coding: utf8 --

r"""
    Parent SAFT EOS class

    An additional class with variant specifics imported from ``saft_type.py``
    completes the EOS.
"""

import numpy as np
import logging

from despasito.equations_of_state import constants
import despasito.utils.general_toolbox as gtb
import despasito.equations_of_state.eos_toolbox as tb
from despasito.equations_of_state.interface import EosTemplate

from despasito.equations_of_state.saft import Aideal
from despasito.equations_of_state.saft import Aassoc

logger = logging.getLogger(__name__)


def saft_type(name):
    r"""
    Initialize EOS object for SAFT variant.

    All input and calculated parameters are defined as hidden attributes.

    Parameters
    ----------
    name : str
        Name of supported saft variant, the following are currently supported:

        - gamma_mie: :class:`~despasito.equations_of_state.saft.gamma_mie.SaftType`
        - gamma_sw: :class:`~despasito.equations_of_state.saft.gamma_sw.SaftType`

    Returns
    -------
    saft_source : obj
        SAFT variant class to be initiated in
        :class:`~despasito.equations_of_state.saft.saft.EosType`

    """
    if name == "gamma_mie":
        from despasito.equations_of_state.saft.gamma_mie import SaftType as saft_source
    elif name == "gamma_sw":
        from despasito.equations_of_state.saft.gamma_sw import SaftType as saft_source
    else:
        raise ValueError(
            "SAFT type, {}, is not supported. Be sure the class".format(name)
            + " is added to the factory function 'saft_type'"
        )

    return saft_source


class EosType(EosTemplate):
    r"""
    Initialize EOS object for SAFT variant.

    All input and calculated parameters are defined as hidden attributes.

    Parameters
    ----------
    beads : list[str]
        List of unique bead names used among components
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters:

        - mass: Bead mass [kg/mol]
        - epsilonHB-\*-\*: Optional, Interaction energy between each bead and
        association site. Asterisk represents string from sitenames.
        - K-\*-\*: Optional, Bonding volume between each association site. Asterisk
        represents two strings from sitenames.
        - rc-\*-\*: Optional, Cutoff distance for association sites. Asterisk
        represents two strings from sitenames.
        - rd-\*-\*: Optional, Site position. Asterisk represents two strings from
        sitenames.
        - Nk-\*: Optional, The number of sites of from list sitenames. Asterisk
        represents string from sitenames.
        - etc. depending on SAFT variant

    cross_library : dict, Optional, default={}
        Optional library of bead cross interaction parameters. As many or as few of
        the desired parameters may be defined for whichever group combinations are
        desired.
        Note that the astericks below represent first the sitename of the primary bead
        and second the sitename of the secondary bead in the .json or dictionary
        structure.

        - epsilonHB-\*-\*: Optional, Interaction energy between each bead and
        association site. Asterisk represents string from sitenames.
        - K-\*-\*: Optional, Bonding volume between each association site. Asterisk
        represents two strings from sitenames.
        - rc-\*-\*: Optional, Cutoff distance for association sites. Asterisk
        represents two strings from sitenames.
        - rd-\*-\*: Optional, Site position. Asterisk represents two strings from
        sitenames.
        - etc. depending on SAFT variant

    combining_rules : dict, Optional, default=None
        Provided to overwrite functional form of combining rules defined for
        parameters in specific SAFT variant. See appropriate class.
    saft_name : str, Optional, default="gamma_mie"
        Define the SAFT variant, options listed in
        :func:`~despasito.equations_of_state.saft.saft.saft_type`.
    molecular_composition : numpy.ndarray
        :math:`\nu_{i,k}/k_B`. Array of number of components by number of bead types.
        Defines the number of each type of group in each component, as it corresponds
        to the ``beads`` array.
    Aideal_method : str, Optional
        Functional form of ideal gas contribution for Helmholtz energy. Default is
        defined in SAFT variant defined with ``saft_name``. See
        :func:`~despasito.equations_of_state.saft.Aideal.Aideal_contribution` for more
        options.
    reduction_ratio : float, Optional
        Reduced distance of the sites from the center of the sphere of interaction.
        This value is used when site position, ``eos_dict['rd_klab'] == None``. See
        :func:`~despasito.equations_of_state.saft.Aassoc.calc_bonding_volume` for more
        details.
    method_stat : obj
        EOS object containing the the method status of the available options.
    kwargs
        Other keywords that are specific to the chosen SAFT variant

    Attributes
    ----------
    beads : list[str]
        List of unique bead names used among components
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters. See entry in **Parameters** section.
    cross_library : dict
        Library of bead cross interaction parameters. As many or as few of the desired
        parameters may be defined for whichever group combinations are desired. See
        entry in **Parameters** section.
    number_of_components : int
        Number of components in mixture represented by given EOS object.
    parameter_types : list[str]
        This list of parameter names for the association site calculation, "epsilonHB",
        "K", "rc", and/or "rd" as well as parameters for the specific SAFT variant.
    parameter_bound_extreme : dict
        With each parameter names as an entry representing a list with the minimum and
        maximum feasible parameter value.

        - epsilonHB: [100.,5000.]
        - K: [1e-5,10000.]
        - rc: [0.1,10.0]
        - rd: [0.1,10.0]
        - etc., other parameters from SAFT variant

    saft_name : str, Optional, default="gamma_mie"
        Define the SAFT variant, options listed in
        :func:`~despasito.equations_of_state.saft.saft.saft_type`.
    saft_source : obj
        Object representing SAFT variant. This attribute can be used to access
        intermediate calculations.
    eos_dict : dict
        Temperature value is initially defined as NaN for a placeholder until
        temperature dependent attributes are initialized by using a method of this
        class.

        - molecular_composition (numpy.ndarray) - :math:`\nu_{i,k}/k_B`. Array
        containing the number of components by the number of bead types. Defines the
        number of each type of group in each component
        - residual_helmholtz_contributions (list[str]) - List of methods from the
        specified saft_source representing contributions to the Helmholtz energy that
        are functions of density, temperature, and composition
        - Aideal_method (str) - Method defined in SAFT variant
        - massi (list) - List of molecular weight for each group in ``beads`` array
        - flag_assoc (bool) - flag indicating whether there is an association site
        contribution to the Helmholtz energy
        - sitenames (list[str]) - List of sitenames for association site interactions.
        This array is extracted from ``bead_library`` entries
        - nk (numpy.ndarray) - A matrix of (Nbeads x Nsites) Contains for each bead the
        number of each type of site
        - epsilonHB (numpy.ndarray) - Optional, Interaction energy between each bead and
        association site.
        - Kklab (numpy.ndarray) - Optional, Bonding volume between each association site
        - rc_klab, Optional, Cutoff distance for association sites
        - rd_klab, Optional, Association site position
        - reduction_ratio (float) - Reduced distance of the sites from the center of the
        sphere of interaction. This value is used when site position,
        ``eos_dict['rd_klab'] == None``.

    """

    def __init__(self, saft_name="gamma_mie", Aideal_method=None, combining_rules=None, **kwargs):

        super().__init__(**kwargs)

        self.saft_name = saft_name
        saft_source = saft_type(saft_name)
        self.saft_source = saft_source(**kwargs)

        if "method_stat" in kwargs:
            self.method_stat = self.saft_source.method_stat

        if not hasattr(self, "eos_dict"):
            self.eos_dict = {}

        # Extract needed variables from saft type file (e.g. gamma_mie)
        self.parameter_types = ["epsilonHB", "K", "rc", "rd"]
        self.parameter_bound_extreme = {
            "epsilonHB": [100.0, 5000.0],
            "K": [1e-5, 10000.0],
            "rc": [0.1, 10.0],
            "rd": [0.1, 10.0],
        }

        saft_attributes = [
            "Aideal_method",
            "parameter_types",
            "parameter_bound_extreme",
            "residual_helmholtz_contributions",
        ]
        for key in saft_attributes:
            try:
                tmp = getattr(self.saft_source, key)
                if key == "parameter_bound_extreme":
                    self.parameter_bound_extreme.update(tmp)
                elif key == "parameter_types":
                    self.parameter_types = list(set(self.parameter_types) | set(tmp))
                else:
                    self.eos_dict[key] = tmp
            except Exception:
                raise ValueError("SAFT type, {}, is missing the variable {}.".format(saft_name, key))

        for res in self.eos_dict["residual_helmholtz_contributions"]:
            setattr(self, res, getattr(self.saft_source, res))

        if Aideal_method is not None:
            logger.info("Switching Aideal method from {} to {}.".format(self.eos_dict["Aideal_method"], Aideal_method))
            self.eos_dict["Aideal_method"] = Aideal_method

        # Extract needed values from kwargs
        needed_attributes = ["bead_library", "molecular_composition", "beads"]
        for key in needed_attributes:
            if key not in kwargs:
                raise ValueError("The one of the following inputs is missing: {}".format(", ".join(tmp)))

            if key == "molecular_composition":
                self.eos_dict[key] = kwargs[key]
            else:
                setattr(self, key, kwargs[key])
        self.number_of_components = len(self.eos_dict["molecular_composition"])

        if "cross_library" not in kwargs:
            self.cross_library = {}
        else:
            self.cross_library = kwargs["cross_library"]

        if "massi" not in self.eos_dict:
            self.eos_dict["massi"] = tb.calc_massi(
                self.eos_dict["molecular_composition"], self.bead_library, self.beads
            )

        if "reduction_ratio" in kwargs:
            self.eos_dict["reduction_ratio"] = kwargs["reduction_ratio"]

        # Initiate association site terms
        self.eos_dict["sitenames"], self.eos_dict["nk"], self.eos_dict["flag_assoc"] = Aassoc.initiate_assoc_matrices(
            self.beads, self.bead_library, self.eos_dict["molecular_composition"]
        )
        assoc_output = Aassoc.calc_assoc_matrices(
            self.beads,
            self.bead_library,
            self.eos_dict["molecular_composition"],
            sitenames=self.eos_dict["sitenames"],
            cross_library=self.cross_library,
            nk=self.eos_dict["nk"],
        )
        self.eos_dict.update(assoc_output)
        if np.size(np.where(self.eos_dict["epsilonHB"] != 0.0)) == 0:
            self.eos_dict["flag_assoc"] = False

        if self.eos_dict["flag_assoc"]:
            self.T = None

        if combining_rules is not None:
            logger.info("Accepted new combining rule definitions")
            self.saft_source.combining_rules.update(combining_rules)
            self.parameter_refresh()

    def residual_helmholtz_energy(self, rho, T, xi):
        r"""
        Return a vector of residual Helmholtz energy.

        :math:`\frac{A^{res}}{N k_{B} T}`

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0


        Returns
        -------
        Ares : numpy.ndarray
            Residual Helmholtz energy for each density value given.
        """
        if len(xi) != self.number_of_components:
            raise ValueError(
                "Number of components in mole fraction list, {}, ".format(len(xi))
                + "doesn't match self.number_of_components, {}".format(self.number_of_components)
            )

        rho = self._check_density(rho)

        if any(np.array(xi) < 0.0):
            raise ValueError("Mole fractions cannot be less than zero.")

        Ares = np.zeros(len(rho))
        for res in self.eos_dict["residual_helmholtz_contributions"]:
            Ares += getattr(self.saft_source, res)(rho, T, xi)

        if self.eos_dict["flag_assoc"]:
            Ares += self.Aassoc(rho, T, xi)

        return Ares

    def helmholtz_energy(self, rho, T, xi):
        r"""
        Return a vector of Helmholtz energy.

        :math:`\frac{A}{N k_{B} T}`

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Returns
        -------
        A : numpy.ndarray
            Total Helmholtz energy for each density value given.
        """
        if len(xi) != self.number_of_components:
            raise ValueError(
                "Number of components in mole fraction list, {}, ".format(len(xi))
                + "doesn't match self.number_of_components, {}".format(self.number_of_components)
            )

        rho = self._check_density(rho)

        A = self.residual_helmholtz_energy(rho, T, xi) + self.Aideal(rho, T, xi, method=self.eos_dict["Aideal_method"])

        return A

    def Aideal(self, rho, T, xi, method="Abroglie"):
        r"""
        Return a vector of ideal contribution of Helmholtz energy.

        :math:`\frac{A^{ideal}}{N k_{B} T}`

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0
        method : str, Optional, default=Abroglie
            The function name of the method to calculate the ideal contribution of
            the Helmholtz energy. To add a new one, add a function to:
            despasito.equations_of_state.saft.Aideal.py

        Returns
        -------
        Aideal : numpy.ndarray
            Helmholtz energy of ideal gas for each density given.

        """
        if len(xi) != self.number_of_components:
            raise ValueError(
                "Number of components in mole fraction list, {}, ".format(len(xi))
                + "doesn't match self.number_of_components, {}".format(self.number_of_components)
            )

        rho = self._check_density(rho)

        return Aideal.Aideal_contribution(rho, T, xi, self.eos_dict["massi"], method=method)

    def Aassoc(self, rho, T, xi):
        r"""
        Return a vector of association site contribution of Helmholtz energy.

        :math:`\frac{A^{association}}{N k_{B} T}`

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : numpy.ndarray
            Mole fraction of each component, sum(xi) should equal 1.0

        Returns
        -------
        Aassoc : numpy.ndarray
            Helmholtz energy of ideal gas for each density given.
        """

        if self.eos_dict["flag_assoc"]:
            if len(xi) != self.number_of_components:
                raise ValueError(
                    "Number of components in mole fraction list, {}, ".format(len(xi))
                    + "doesn't match self.number_of_components, {}".format(self.number_of_components)
                )

            rho = self._check_density(rho)

            indices = Aassoc.assoc_site_indices(self.eos_dict["nk"], self.eos_dict["molecular_composition"], xi=xi)

            # compute F_klab
            Fklab = np.exp(self.eos_dict["epsilonHB"] / T) - 1.0
            if "rc_klab" in self.eos_dict:
                if "Kijklab" not in self.eos_dict or T != self.T:
                    opts = {}
                    keys = ["rd_klab", "reduction_ratio"]
                    for key in keys:
                        if key in self.eos_dict:
                            opts[key] = self.eos_dict[key]
                    self.eos_dict["Kijklab"] = self.saft_source.calc_Kijklab(T, self.eos_dict["rc_klab"], **opts)
                    self.T = T
                Kklab = self.eos_dict["Kijklab"]
                Ktype = "ijklab"
            else:
                Kklab = self.eos_dict["Kklab"]
                Ktype = "klab"

            gr_assoc = self.saft_source.calc_gr_assoc(rho, T, xi, Ktype=Ktype)

            # Compute Xika: with python with numba  {BottleNeck}

            Xika = Aassoc._calc_Xika_wrap(
                indices,
                rho,
                xi,
                self.eos_dict["molecular_composition"],
                self.eos_dict["nk"],
                Fklab,
                Kklab,
                gr_assoc,
                method_stat=self.method_stat,
            )

            # Compute A_assoc
            Assoc_contribution = np.zeros(np.size(rho))
            for ind, (i, k, a) in enumerate(indices):
                if self.eos_dict["nk"][k, a] != 0.0:
                    tmp = np.log(Xika[:, ind]) + ((1.0 - Xika[:, ind]) / 2.0)
                    Assoc_contribution += (
                        xi[i] * self.eos_dict["molecular_composition"][i, k] * self.eos_dict["nk"][k, a] * tmp
                    )

        else:
            logger.warning(
                "Association Site contribution was called when the appropriate "
                "parameters were not provided. This should not occur."
            )
            Assoc_contribution = None

        return Assoc_contribution

    def pressure(self, rho, T, xi, step_size=1e-6):
        """
        Compute pressure given system information.

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : list[float]
            Mole fraction of each component, sum(xi) should equal 1.0

        Returns
        -------
        P : numpy.ndarray
            Array of pressure values [Pa] associated with each density and so equal in
            length
        """
        if len(xi) != self.number_of_components:
            raise ValueError(
                "Number of components in mole fraction list, {}, ".format(len(xi))
                + "doesn't match self.number_of_components, {}".format(self.number_of_components)
            )

        # derivative of Aideal_broglie here wrt to rho is 1/rho
        rho = self._check_density(rho)
        P_tmp = gtb.central_difference(rho, self.helmholtz_energy, args=(T, xi), step_size=step_size, relative=True)
        pressure = P_tmp * T * constants.R * rho**2

        return pressure

    def fugacity_coefficient(self, P, rho, xi, T, dy=1e-5, log_method=True):
        """
        Compute fugacity coefficient.

        Parameters
        ----------
        P : float
            Pressure of the system [Pa]
        rho : float
            Molar density of system [:math:`mol/m^3`]
        T : float
            Temperature of the system [K]
        xi : list[float]
            Mole fraction of each component, sum(xi) should equal 1.0
        log_method : bool, Optional, default=False
            Choose to use a log transform in central difference method. This allows
            easier calculations for very small numbers.

        Returns
        -------
        fugacity_coefficient : numpy.ndarray
            Array of fugacity coefficient values for each component
        """
        if len(xi) != self.number_of_components:
            raise ValueError(
                "Number of components in mole fraction list, {}, ".format(len(xi))
                + "doesn't match self.number_of_components, {}".format(self.number_of_components)
            )

        if gtb.isiterable(T):
            if len(T) == 1:
                T = T[0]
            else:
                raise ValueError("Temperature must be given as a scalar.")
        if gtb.isiterable(rho):
            if len(rho) == 1:
                rho = rho[0]
            else:
                raise ValueError("Density must be given as a scalar.")
        if gtb.isiterable(P):
            if len(P) == 1:
                P = P[0]
            else:
                raise ValueError("Pressure must be given as a scalar.")

        rho = self._check_density(rho)
        logZ = np.log(P / (rho * T * constants.R))
        Ares = self.residual_helmholtz_energy(rho, T, xi)
        dAresdrho = tb.partial_density_central_difference(
            xi, rho, T, self.residual_helmholtz_energy, step_size=dy, log_method=True
        )

        phi = np.exp(Ares + rho * dAresdrho - logZ)

        return phi

    def density_max(self, xi, T, maxpack=0.65):
        """
        Estimate the maximum density based on the hard sphere packing fraction.

        Parameters
        ----------
        xi : list[float]
            Mole fraction of each component, sum(xi) should equal 1.0
        T : float
            Temperature of the system [K]
        maxpack : float, Optional, default=0.65
            Maximum packing fraction

        Returns
        -------
        max_density : float
            Maximum molar density [:math:`mol/m^3`]
        """
        if len(xi) != self.number_of_components:
            raise ValueError(
                "Number of components in mole fraction list, {}, ".format(len(xi))
                + "doesn't match self.number_of_components, {}".format(self.number_of_components)
            )

        max_density = self.saft_source.density_max(xi, T, maxpack=maxpack)

        return max_density

    def check_bounds(self, parameter, param_name, bounds):
        """
        Ensures that bounds given for parameter are within the range of feasibility
        defined in this class. If the bounds are outside of the allowed range, they
        are adjusted.

        Parameters
        ----------
        parameter : str
            Parameter type to be fit (e.g. for 'epsilon_CO2', this argument will be
            'epsilon').
        param_name : str
            Full parameter string to be fit. See EOS documentation for supported
            parameter names.
        bounds : list
            Upper and lower bound for given parameter type

        Returns
        -------
        bounds : list
            A screened and possibly corrected low and a high value for the parameter,
            param_name
        """

        # Remove association site names
        param_name = param_name.split("-")[0]
        bounds_new = super().check_bounds(parameter, param_name, bounds)

        return bounds_new

    def update_parameter(self, param_name, bead_names, param_value):
        r"""
        Update a single parameter value during parameter fitting process.

        To refresh those parameters that are dependent on to bead_library or
        cross_library, use method ``parameter_refresh``.

        Parameters
        ----------
        param_name : str
            Parameter to be fit. See EOS documentation for supported parameter names.
            Cross interaction parameter names should be composed of parameter name and
            the other bead type, separated by an underscore (e.g. epsilon_CO2).
        bead_names : list
            Bead names to be changed. For a self interaction parameter, the length
            will be one, for a cross interaction parameter, the length will be two.
        param_value : float
            Value of parameter
        """

        parameter_list = param_name.split("-")
        if len(parameter_list) > 1 and len(parameter_list[1:]) != 2:
            raise ValueError(
                "Sitenames should be two different sites in the list:"
                + " {}. You gave: {}".format(self.eos_dict["sitenames"], ", ".join(parameter_list[1:]))
            )

        super().update_parameter(param_name, bead_names, param_value)

    def parameter_refresh(self):
        r"""
        To refresh dependent parameters

        Those parameters that are dependent on bead_library and cross_library attributes
        **must** be updated by running this function after all parameters have been
        adjusted with ``update_parameters`` method.

        """

        self.saft_source.parameter_refresh(self.bead_library, self.cross_library)

        # Update Association site matrices
        if self.eos_dict["flag_assoc"]:
            assoc_output = Aassoc.calc_assoc_matrices(
                self.beads,
                self.bead_library,
                self.eos_dict["molecular_composition"],
                sitenames=self.eos_dict["sitenames"],
                cross_library=self.cross_library,
                nk=self.eos_dict["nk"],
            )
            self.eos_dict.update(assoc_output)

    def _check_density(self, rho):
        r"""
        This function checks the attributes of the density array

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [:math:`mol/m^3`]
        """

        if np.isscalar(rho):
            rho = np.array([rho])
        elif not isinstance(rho, np.ndarray):
            rho = np.array(rho)
        if len(np.shape(rho)) == 2:
            rho = rho[0]

        if any(np.isnan(rho)):
            raise ValueError("NaN was given as a value of density, rho")
        elif rho.size == 0:
            raise ValueError("No value of density, rho, was given")
        elif any(rho < 0.0):
            raise ValueError("Density values cannot be negative.")

        return rho

    def __str__(self):

        string = "EOS: SAFT-{}, Beads: {},\nMasses: {} kg/mol\nSitenames: {}".format(
            self.saft_name,
            self.beads,
            self.eos_dict["massi"],
            self.eos_dict["sitenames"],
        )
        return string
